import abc
import math
from typing import Union, List, Dict, Set, Optional, Tuple
import numpy as np

from commonroad.common.util import Interval
from commonroad.common.validity import is_valid_orientation, is_real_number_vector
from commonroad.geometry.shape import Shape, \
    occupancy_shape_from_state
from commonroad.scenario.trajectory import Trajectory

__author__ = "Stefanie Manzinger"
__copyright__ = "TUM Cyber-Physical Systems Group"
__credits__ = ["Priority Program SPP 1835 Cooperative Interacting Automobiles",
               "CAR@TUM"]
__version__ = "2021.2"
__maintainer__ = "Stefanie Manzinger"
__email__ = "commonroad@lists.lrz.de"
__status__ = "Released"

from commonroad.visualization.drawable import IDrawable
from commonroad.visualization.param_server import ParamServer
from commonroad.visualization.renderer import IRenderer


class Occupancy(IDrawable):
    """ Class describing an occupied area in the position domain. The
    occupied area can be defined for a certain time
    step or a time interval."""

    def __init__(self, time_step: Union[int, Interval], shape: Shape):
        """
        :param time_step: a time interval or time step for which the
        occupancy is defined
        :param shape: occupied region in the position domain
        """
        self.time_step: Union[int, Interval] = time_step
        self.shape: Shape = shape

    @property
    def shape(self) -> Shape:
        """ Shape representing an occupied area in the position domain."""
        return self._shape

    @shape.setter
    def shape(self, shape: Shape):
        assert isinstance(shape, Shape), '<Occupancy/shape>: argument "shape" of wrong type. Expected type: %s. ' \
                                         'Got type: %s.' % (Shape, type(shape))
        self._shape = shape

    @property
    def time_step(self) -> Union[int, Interval]:
        """ The occupied area is either defined for a certain time step or a time interval."""
        return self._time_step

    @time_step.setter
    def time_step(self, time_step: Union[int, Interval]):
        assert isinstance(time_step, (int, Interval)), '<Occupancy/time_step>: argument "time_step" of ' \
                                                       'wrong type. Expected type: %s or %s. Got type: %s.' \
                                                       % (int, Interval, type(time_step))
        self._time_step = time_step

    def translate_rotate(self, translation: np.ndarray, angle: float):
        """ Translates and rotates the occupied area.

        :param translation: translation vector [x_off, y_off] in x- and
        y-direction
        :param angle: rotation angle in radian (counter-clockwise)
        """
        assert is_real_number_vector(translation,
                                     2), '<Occupancy/translate_rotate>: ' \
                                         'argument "translation" is ' \
                                         'not a vector of real numbers of ' \
                                         'length 2.'
        assert is_valid_orientation(
                angle), '<Occupancy/translate_rotate>: argument "orientation" ' \
                        'is ' \
                        'not valid.'

        self._shape = self._shape.translate_rotate(translation, angle)

    def draw(self, renderer: IRenderer,
             draw_params: Union[ParamServer, dict, None] = None,
             call_stack: Optional[Tuple[str, ...]] = tuple()):
        call_stack = tuple(list(call_stack) + ['occupancy'])
        self.shape.draw(renderer, draw_params, call_stack)


class Prediction:
    """
        Base class for a prediction module.
    """
    def __init__(self, initial_time_step: int, occupancy_set: List[Occupancy]):
        """
        :param initial_time_step: initial time step of the prediction
        :param occupancy_set: list of occupancies defined for different time steps or time intervals.
        """
        self.initial_time_step: int = initial_time_step
        self.occupancy_set: List[Occupancy] = occupancy_set

    @property
    def initial_time_step(self) -> int:
        """ Initial time step of the prediction."""
        return self._initial_time_step

    @initial_time_step.setter
    def initial_time_step(self, initial_time_step: int):
        assert isinstance(initial_time_step, int), '<Prediction/initial_time_step>: argument "initial_time_step" of ' \
                                                   'wrong type. Expected type: %s. Got type: %s.' \
                                                   % (int, type(initial_time_step))
        self._initial_time_step = initial_time_step

    @property
    def final_time_step(self) -> Union[int, Interval]:
        """ Final time step of the prediction."""
        return self._final_time_step

    @final_time_step.setter
    def final_time_step(self, final_time_step: Union[int, Interval]):
        assert isinstance(final_time_step, (int, Interval)), '<Prediction/final_time_step>: argument "final_time_step" of ' \
                                                   'wrong type. Expected type: %s. Got type: %s.' \
                                                   % ([int, Interval], type(final_time_step))
        self._final_time_step = final_time_step

    @property
    def occupancy_set(self) -> List[Occupancy]:
        """ List of occupancies over time."""
        return self._occupancy_set

    @occupancy_set.setter
    def occupancy_set(self, occupancy_set: List[Occupancy]):
        assert isinstance(occupancy_set, list), '<Prediction/occupancy_set>: argument "occupancy_set" of wrong type. ' \
                                                'Expected type: %s. Got type: %s.' % (list, type(occupancy_set))
        assert all(isinstance(occupancy, Occupancy) for occupancy in occupancy_set), '<Prediction/occupancy_set>: ' \
                                                                                     'element of "occupancy_set" is ' \
                                                                                     'of wrong type. Expected type: ' \
                                                                                     '%s.' % Occupancy
        self._occupancy_set = occupancy_set
        self.final_time_step = max([occ.time_step for occ in self._occupancy_set])

    def occupancy_at_time_step(self, time_step: int) -> Union[None, Occupancy]:
        """ Occupancy at a specific time step.

        :param time_step: discrete time step
        :return: occupancy at time_step if time_step is within the time interval of the prediction; otherwise, None
        """
        assert isinstance(time_step, int), '<Prediction/occupancy_at_time_step>: argument "time_step" of ' \
                                           'wrong type. Expected type: %s. Got type: %s.' % (int, type(time_step))
        occupancy = None
        for occ in self._occupancy_set:
            if isinstance(occ.time_step, Interval):
                if occ.time_step.contains(time_step):
                    occupancy = occ
                    break
            elif isinstance(occ.time_step, int):
                if occ.time_step == time_step:
                    occupancy = occ
                    break
        return occupancy

    @abc.abstractmethod
    def translate_rotate(self, translation: list, angle: float):
        pass


class SetBasedPrediction(Prediction):
    """ Class to represent the future behavior of obstacles by bounded occupancy sets."""
    def __init__(self, initial_time_step: int, occupancy_set: List[Occupancy]):
        """
        :param initial_time_step: initial time step of the set-based prediction
        :param occupancy_set: list of occupancies defined for different time steps or time intervals.
        """
        Prediction.__init__(self, initial_time_step, occupancy_set)

    def translate_rotate(self, translation: np.ndarray, angle: float):
        """ Translates and rotates the occupancy set.

        :param translation: translation vector [x_off, y_off] in x- and y-direction
        :param angle: rotation angle in radian (counter-clockwise)
        """
        assert is_real_number_vector(translation, 2), '<SetBasedPrediction/translate_rotate>: argument "translation" ' \
                                                      'is not a vector of real numbers of length 2.'
        assert is_valid_orientation(angle), '<SetBasedPrediction/translate_rotate>: argument "orientation" ' \
                                            'is not valid.'
        for occ in self._occupancy_set:
            occ.translate_rotate(translation, angle)


class TrajectoryPrediction(Prediction):
    """ Class to represent the predicted movement of an obstacle using a trajectory. A trajectory is modeled as a
    state sequence over time. The occupancy of an obstacle along a trajectory is uniquely defined given its shape."""
    def __init__(self, trajectory: Trajectory, shape: Shape,
                 center_lanelet_assignment: Union[None, Dict[int, Set[int]]] = None,
                 shape_lanelet_assignment: Union[None, Dict[int, Set[int]]] = None):
        """
        :param trajectory: predicted trajectory of the obstacle
        :param shape: shape of the obstacle
        """
        self.shape: Shape = shape
        self.trajectory: Trajectory = trajectory
        self.shape_lanelet_assignment: Dict[int, Set[int]] = shape_lanelet_assignment
        self.center_lanelet_assignment: Dict[int, Set[int]] = center_lanelet_assignment
        Prediction.__init__(self, self._trajectory.initial_time_step, self._create_occupancy_set())

    @property
    def shape(self) -> Shape:
        """ Shape of the predicted object."""
        return self._shape

    @shape.setter
    def shape(self, shape: Shape):
        assert isinstance(shape, Shape), '<TrajectoryPrediction/shape>: argument "shape" of wrong type. Expected ' \
                                         'type: %s. Got type: %s.' % (Shape, type(shape))
        self._shape = shape

    @property
    def trajectory(self) -> Trajectory:
        """ Predicted trajectory of the object."""
        return self._trajectory

    @trajectory.setter
    def trajectory(self, trajectory: Trajectory):
        assert isinstance(trajectory, Trajectory), '<TrajectoryPrediction/trajectory>: argument "trajectory" of wrong' \
                                                   ' type. Expected type: %s. Got type: %s.' \
                                                   % (Trajectory, type(trajectory))
        self._trajectory = trajectory

    @property
    def shape_lanelet_assignment(self) -> Union[None, Dict[int, Set[int]]]:
        """ Predicted lanelet assignment of obstacle shape."""
        return self._shape_lanelet_assignment

    @shape_lanelet_assignment.setter
    def shape_lanelet_assignment(self, shape_lanelet_assignment: Union[None, Dict[int, Set[int]]]):
        if shape_lanelet_assignment is not None:
            assert isinstance(shape_lanelet_assignment, dict), '<TrajectoryPrediction/shape_lanelet_assignment>: ' \
                                                         'argument "shape_lanelet_assignment" of wrong type. ' \
                                                               'Expected type: %s. Got' \
                                                               ' type: %s.' % (Dict, type(shape_lanelet_assignment))
        self._shape_lanelet_assignment = shape_lanelet_assignment

    @property
    def center_lanelet_assignment(self) -> Union[None, Dict[int, Set[int]]]:
        """ Predicted lanelet assignment of obstacle center."""
        return self._center_lanelet_assignment

    @center_lanelet_assignment.setter
    def center_lanelet_assignment(self, center_lanelet_assignment: Union[None, Dict[int, Set[int]]]):
        if center_lanelet_assignment is not None:
            assert isinstance(center_lanelet_assignment, dict), '<TrajectoryPrediction/center_lanelet_assignment>: ' \
                                                         'argument "center_lanelet_assignment" of wrong type. ' \
                                                         'Expected type: ' \
                                                         '%s. Got type: %s.' % (Dict, type(center_lanelet_assignment))
        self._center_lanelet_assignment = center_lanelet_assignment

    def translate_rotate(self, translation: np.ndarray, angle: float):
        """ Translates and rotates all states of the trajectory and re-computes the translated and rotated occupancy
        set.

        :param translation: translation vector [x_off, y_off] in x- and y-direction
        :param angle: rotation angle in radian (counter-clockwise)
        """
        assert is_real_number_vector(translation, 2), '<TrajectoryPrediction/translate_rotate>: argument ' \
                                                      '"translation" is not a vector of real numbers of length 2.'
        assert is_valid_orientation(angle), '<TrajectoryPrediction/translate_rotate>: argument "orientation" is ' \
                                            'not valid.'

        self._trajectory.translate_rotate(translation, angle)
        self._occupancy_set = self._create_occupancy_set()

    def _create_occupancy_set(self):
        """ Computes the occupancy set over time given the predicted trajectory and shape of the object."""
        occupancy_set = list()
        for k, state in enumerate(self._trajectory.state_list):
            if not hasattr(state, "orientation"):
                state.orientation = math.atan2(state.velocity_y, state.velocity)
            occupied_region = occupancy_shape_from_state(self._shape, state)
            occupancy_set.append(Occupancy(state.time_step, occupied_region))
        return occupancy_set



from abc import ABC, abstractmethod
from typing import Union, Tuple, Optional

from commonroad.visualization.param_server import ParamServer
from commonroad.visualization.renderer import IRenderer


class IDrawable(ABC):
    """
    Interface for drawable types
    """

    @abstractmethod
    def draw(self, renderer: IRenderer, draw_params: Union[ParamServer, dict, None],
             call_stack: Optional[Tuple[str, ...]]) -> None:
        """
        Draw the object

        :param renderer: Renderer to use for drawing
        :param draw_params: Optional parameters ovrriding the defaults for plotting given by a nested dict that
            recreates the structure of an object or a ParamServer object
        :param call_stack: Optional tuple of string containing the call stack, which allows for differentiation of
            plotting styles depending on the call stack
        :return: None
        """
        pass



from typing import List, Union

import matplotlib.pyplot as plt
from matplotlib.animation import FuncAnimation

from commonroad.visualization.draw_dispatch_cr import draw_object, plottable_types
from commonroad.visualization.util import approximate_bounding_box_dyn_obstacles


def create_scenario_video(obj_lists: Union[List[plottable_types]], file_path: str, time_begin: int, time_end: int,
                          delta_time_steps: int = 1, plotting_horizon=0, plot_limits: Union[list, None, str] = None,
                          draw_params: Union[dict, None] = {}, fig_size: Union[list, None] = None, dt=0.1, dpi=120):
    """
    Creates a video of one or multiple CommonRoad objects in mp4, gif,
    or avi format.

    :param obj: list of objects to be plotted. When plotting objects of the
        same type, use list of lists for speed-up.
    :param file_path: filename of generated video (ends on .mp4/.gif/.avi, default mp4, when nothing is specified)
    :param time_begin: first time step of video
    :param time_end: last time step of video
    :param delta_time_steps: plot every delta_time_steps time steps of scenario
    :param plotting_horizon: time steps of prediction plotted in each frame
    :param plot_limits: axis limits or 'auto' for limiting view to dynamic vehicles
    :param draw_params: draw_params (see draw_object doc)
    :param fig_size: size of the video
    :param dt: time step between frames frames
    :param dpi: resolution of the video
    :return: None
    """
    assert time_begin < time_end, '<video/create_scenario_video> ' \
                                  'time_begin=%i needs to smaller than ' \
                                  'time_end=%i.' % (time_begin, time_end)

    if fig_size is None:
        fig_size = [15, 8]

    fig = plt.figure(figsize=(fig_size[0], fig_size[1]))
    ln, = plt.plot([], [], animated=True)

    def update(frame=0):
        # plot frame
        plt.clf()
        plt.gca().set_aspect('equal')
        ax = plt.gca()
        draw_params.update({
                'time_begin': time_begin + delta_time_steps * frame,
                'time_end':   time_begin + min(frame_count,
                                               delta_time_steps * frame +
                                               plotting_horizon)
        })
        plot_limits_tmp = None if plot_limits == 'auto' else plot_limits
        for obj in obj_lists:
            draw_object(obj, ax=ax, draw_params=draw_params,
                        plot_limits=plot_limits_tmp)
        # Set limits to assure that each frame has the same size
        if plot_limits is None:
            ax.autoscale()
        elif plot_limits == 'auto':
            limits = approximate_bounding_box_dyn_obstacles(obj_lists, frame)
            if limits is not None:
                plt.xlim(limits[0][0] - 10, limits[0][1] + 10)
                plt.ylim(limits[1][0] - 10, limits[1][1] + 10)
            else:
                ax.autoscale()

        return ln,

    frame_count = (time_end - time_begin) // delta_time_steps
    # Interval determines the duration of each frame
    interval = 1.0 / dt
    plt.ioff()
    anim = FuncAnimation(fig, update, frames=frame_count, init_func=update,
                         blit=False, interval=interval)

    if not any([file_path.endswith('.mp4'), file_path.endswith('.gif'),
                file_path.endswith('.avi')]):
        file_path += '.mp4'

    anim.save(file_path, dpi=dpi, writer='ffmpeg')

    plt.close(plt.gcf())



import numpy as np
import math
from scipy.spatial.transform.rotation import Rotation
from scipy.spatial.transform import Slerp
from typing import Union, Tuple
from commonroad.common import validity
from commonroad.common.validity import *

__author__ = "Stefanie Manzinger, Moritz Klischat"
__copyright__ = "TUM Cyber-Physical Systems Group"
__credits__ = ["Priority Program SPP 1835 Cooperative Interacting Automobiles"]
__version__ = "2021.2"
__maintainer__ = "Moritz Klischat"
__email__ = "commonroad@lists.lrz.de"
__status__ = "Released"


def interpolate_angle(x: Union[float, np.array], xp: np.array, fp: np.array, degrees=False):
    """
    :param x: The x-coordinates at which to evaluate the interpolated values.
    :param xp: The x-coordinates of the data points.
    :param fp: The y-coordinates (angles) of the data points, same length as xp.
    :param degrees: True if the input and returned angles are in degrees
    :return: The interpolated angles in radian, same shape as x.
    """

    rotations = Rotation.from_euler("z", fp, degrees=degrees)
    slerp = Slerp(xp, rotations)
    return slerp(x).as_euler("zxy", degrees=degrees)[0]


def subtract_orientations(lhs, rhs):
    """Return the signed difference between angles lhs and rhs
    :param lhs: lhs of the subtraction
    :param rhs: rhs of the subtraction
    :return: ``(lhs - rhs)``, the value will be within ``[-math.pi, math.pi)``.
    Both ``lhs`` and ``rhs`` may either be zero-based (within
    ``[0, 2*math.pi]``), or ``-pi``-based (within ``[-math.pi, math.pi]``).
    """

    return math.fmod((lhs - rhs) + math.pi * 3.0, 2.0 * math.pi) - math.pi


def make_valid_orientation(angle: float) -> float:
    while angle > TWO_PI:
        angle = angle - TWO_PI
    while angle < -TWO_PI:
        angle = angle + TWO_PI
    return angle


def make_valid_orientation_interval(angle_start: float, angle_end: float) -> Tuple[float, float]:
    while angle_start > TWO_PI or angle_end > TWO_PI:
        angle_start -= TWO_PI
        angle_end -= TWO_PI
    while angle_start < -TWO_PI or angle_start < -TWO_PI:
        angle_start += TWO_PI
        angle_end += TWO_PI
    return angle_start, angle_end


class Interval:
    def __init__(self, start: Union[int, float], end: Union[int, float]):
        self._start = None
        self._end = None
        self.start = start
        self.end = end

    def __iter__(self):
        yield self._start
        yield self._end

    def __truediv__(self, other) -> 'Interval':
        if other > 0.0:
            return type(self)(self._start/other, self._end/other)
        else:
            return type(self)(self._end / other, self._start / other)

    def __mul__(self, other: Union[int,float]) -> 'Interval':
        if other > 0.0:
            return type(self)(self._start * other, self._end * other)
        else:
            return type(self)(self._end * other, self._start * other)

    def __round__(self, n=None) -> 'Interval':
        return type(self)(round(self._start, n), round(self._end, n))

    def __add__(self, other: Union[float, int]) -> 'Interval':
        return type(self)(self._start + other, self._end + other)

    def __sub__(self, other: Union[float, int]) -> 'Interval':
        return type(self)(self._start - other, self._end - other)

    @property
    def start(self) -> Union[int, float]:
        return self._start

    @start.setter
    def start(self, start: Union[int, float]):
        if self._end is not None:
            assert start <= self._end, '<common.util/Interval> start of interval must be <= end, but start {} > end {}'.format(start,self._end)
        self._start = start

    @property
    def end(self) -> Union[int, float]:
        return self._end

    @end.setter
    def end(self, end: Union[int, float]):
        if self._start is not None:
            assert end >= self._start, '<common.util/Interval> start of interval must be <= end, but start {} > end {}'.format(self._start, end)
        self._end = end

    def contains(self, other: Union[float, 'Interval']) -> bool:
        if type(other) is Interval:
            return self.start <= other.start and other.end <= self.end
        else:
            return self.start <= other <= self.end

    def overlaps(self, interval: 'Interval') -> bool:
        return self.end >= interval.start and interval.end >= self.start

    def intersection(self, other: 'Interval') -> Union[None, 'Interval']:
        """
        Returns the intersections with another interval.
        :param other: Interval
        :return: Interval or None if intersection is empty
        """
        if not self.overlaps(other):
            return None
        else:
            return Interval(max(self._start, other._start), min(self._end, other._end))

    @property
    def length(self):
        return self.end - self.start

    def __str__(self):
        info_str = "Interval\n"
        info_str += "start: {}\n".format(self._start)
        info_str += "end: {}\n".format(self._end)
        return info_str

    def __gt__(self, other):
        if isinstance(other, validity.ValidTypes.NUMBERS):
            return True if self._start > other else False
        else:
            return True if self._start > other.end else False

    def __lt__(self, other):
        if isinstance(other, validity.ValidTypes.NUMBERS):
            return True if self._end < other else False
        else:
            return True if self._end < other.start else False


class AngleInterval(Interval):
    """ Allows only angles from interval [-2pi,2pi]"""
    def __init__(self, start: Union[int, float], end: Union[int, float]):
        start, end = make_valid_orientation_interval(start,end)
        assert end - start < TWO_PI, '<common.util/AngleInterval> Interval must not be |start-end| > 2pi'
        Interval.__init__(self, start, end)

    @property
    def start(self) -> Union[int, float]:
        return self._start

    @start.setter
    def start(self, start: Union[int, float]):
        assert is_valid_orientation(start), '<common.util/AngleInterval> start angle needs to be in interval [-2pi,2pi]'
        if self._end is not None:
            assert start <= self._end, '<common.util/Interval> start of interval must be <= end, but start {} > end {}'.format(start,self._end)
        self._start = start

    @property
    def end(self) -> Union[int, float]:
        return self._end

    @end.setter
    def end(self, end: Union[int, float]):
        assert is_valid_orientation(end), '<common.util/AngleInterval> end angle needs to be in interval [-2pi,2pi]'
        if self._start is not None:
            assert end >= self._start, '<common.util/Interval> start of interval must be <= end, but start {} > end {}'.format(self._start, end)
        self._end = end

    def intersect(self, other: 'Interval'):
        raise NotImplementedError()




import warnings
from typing import List, Dict, Union

import matplotlib as mpl
import matplotlib.pyplot as plt

from commonroad.common.util import Interval
from commonroad.scenario.scenario import Scenario
from commonroad.visualization.draw_dispatch_cr import draw_object

__author__ = "Moritz Klischat"
__copyright__ = "TUM Cyber-Physical Systems Group"
__credits__ = [""]
__version__ = "2021.2"
__maintainer__ = "Moritz Klischat"
__email__ = "commonroad@lists.lrz.de"
__status__ = "Released"

"""
Example for a fast updating of dynamic obstacles with matplotlib:
# 1. set backend for matplotlib (Qt5Agg or TkAgg recommended) directly after 
importing matplotlib for the first time
import matplotlib
matplotlib.use('Qt5Agg')
import os
from commonroad.common.file_reader import CommonRoadFileReader

# 2. example function:
def example_plot():        
    filename = os.getcwd() + '/scenarios/NGSIM/US101/USA_US101-4_1_T-1.xml'
    scenario, planning_problem_set = CommonRoadFileReader(filename).open()
    from commonroad.visualization.plot_helper import *
    
    set_non_blocking()  # ensures interactive plotting is activated
    plt.style.use('classic')
    inch_in_cm = 2.54
    figsize = [30, 8]
    
    fig = plt.figure(figsize=(figsize[0] / inch_in_cm, figsize[1] / inch_in_cm))
    fig.gca().axis('equal')
    handles = {}  # collects handles of obstacle patches, plotted by matplotlib
    
    # inital plot including the lanelet network
    draw_object(scenario, handles=handles)
    fig.canvas.draw()
    plt.gca().autoscale()
    
    # loop where obstacle positions are updated
    for i in range(0, nrun):
        # ...
        # change positions of obstacles
        # ...
        redraw_obstacles(scenario, handles=handles, figure_handle=fig, 
        plot_limits=None, draw_params={'time_begin':i})
    
    print((time.time()-t1)/nrun)
        
"""


def set_non_blocking() -> None:
    """
    Ensures that interactive plotting is enabled for non-blocking plotting.

    :return: None
    """

    plt.ion()
    if not mpl.is_interactive():
        warnings.warn(
                'The current backend of matplotlib does not support '
                'interactive '
                'mode: ' + str(
                        mpl.get_backend()) + '. Select another backend with: '
                                             '\"matplotlib.use(\'TkAgg\')\"',
                UserWarning, stacklevel=3)


def redraw_obstacles(scenario: Scenario,
                     handles: Dict[int, List[mpl.patches.Patch]],
                     figure_handle: mpl.figure.Figure, draw_params=None,
                     plot_limits: Union[List[Union[int, float]], None] = None,
                     draw: bool = True) -> None:
    """
    This function is used for fast updating dynamic obstacles of an already
    drawn plot. Saves on average about 80% time
    compared to a complete plot.
    Deletes all dynamic obstacles which are specified in handles and draws
    dynamic obstacles of a scenario.

    :param scenario: scenario from which obstacle should be plotted
    :param handles: dict of obstacle_ids and corresponding patch handles (generated by draw_object function)
    :param plot_limits: axis limits for plot [x_min, x_max, y_min, y_max]
    :param figure_handle: figure handle of current plot
    :param draw: if True, updates are displayed directly in figure
    :return: None
    """
    # remove dynamic obstacle from current plot
    for handles_i in handles.values():
        for handle in handles_i:
            if handle is not None:
                handle.remove()
    handles.clear()

    # redraw dynamic obstacles
    if plot_limits is not None:
        draw_object(scenario.obstacles_by_position_intervals(
                [Interval(plot_limits[0], plot_limits[1]),
                 Interval(plot_limits[2], plot_limits[3])]),
                draw_params=draw_params, plot_limits=plot_limits,
                handles=handles)
    else:
        draw_object(scenario.obstacles, draw_params=draw_params,
                    plot_limits=plot_limits, handles=handles)

    # update plot
    if draw is True:
        ax = figure_handle.gca()
        for handles_i in handles.values():
            for handle in handles_i:
                if handle is not None:
                    ax.draw_artist(handle)

        if mpl.get_backend() == 'TkAgg':
            figure_handle.canvas.draw()

        elif mpl.get_backend() == 'Qt5Agg':
            figure_handle.canvas.update()
        else:
            try:
                figure_handle.canvas.update()
            except:
                raise Exception(
                        '<plot_helper/redraw_dynamic_obstacles> Backend for '
                        'matplotlib needs to be \'Qt5Agg\' or \'TkAgg\' but is'
                        '\'%s\'' % mpl.get_backend())

        figure_handle.canvas.flush_events()



import math
import os
import platform
import re
import subprocess
from xml.dom import minidom
import numpy as np
import xml.etree.ElementTree as et
from enum import Enum, unique
from typing import List, Tuple, Union, Dict
from datetime import datetime
import vehiclemodels.parameters_vehicle1 as p1
import vehiclemodels.parameters_vehicle2 as p2
import vehiclemodels.parameters_vehicle3 as p3

from commonroad.common.validity import is_real_number, is_positive
from commonroad.geometry.shape import Rectangle
from commonroad.prediction.prediction import TrajectoryPrediction
from commonroad.scenario.obstacle import DynamicObstacle, ObstacleType
from commonroad.scenario.scenario import ScenarioID
from commonroad.scenario.trajectory import State, Trajectory

__author__ = "Murat Üste, Christina Miller, Moritz Klischat"
__copyright__ = "TUM Cyber-Physical Systems Group"
__credits__ = ["BMW CAR@TUM"]
__version__ = "2021.2"
__maintainer__ = "Moritz Klischat"
__email__ = "commonroad@lists.lrz.de"
__status__ = "Released"

from vehiclemodels.parameters_vehicle1 import parameters_vehicle1
from vehiclemodels.parameters_vehicle3 import parameters_vehicle3


class SolutionException(Exception):
    """
    Main exception class for solution related exceptions.
    """
    pass


class StateTypeException(SolutionException):
    """
    Main exception class for StateType related exceptions.
    """
    pass


class SolutionReaderException(Exception):
    """
    Main exception class for solution reader related exceptions.
    """
    pass


@unique
class VehicleType(Enum):
    FORD_ESCORT = 1
    BMW_320i = 2
    VW_VANAGON = 3


vehicle_parameters = {VehicleType.FORD_ESCORT: p1.parameters_vehicle1(),
                      VehicleType.BMW_320i: p2.parameters_vehicle2(),
                      VehicleType.VW_VANAGON: p3.parameters_vehicle3()}


@unique
class VehicleModel(Enum):
    PM = 0
    ST = 1
    KS = 2
    MB = 3


@unique
class CostFunction(Enum):
    JB1 = 0
    SA1 = 1
    WX1 = 2
    SM1 = 3
    SM2 = 4
    SM3 = 5
    MW1 = 6
    TR1 = 7


@unique
class StateFields(Enum):
    """
    State Fields enum class for defining the state fields for vehicle models for different trajectory types.

    PM | ST | KS | MB -> Corresponding state fields for trajectory states
    Input             -> Input fields for ST, KS, and MB vehicle models
    PMInput           -> Input fields for PM vehicle model.

    Note: If you change the order of field names, don't forget to change the order on the XMLStateFields enum as well,
    because the indexes have to match.
    """
    PM = ['position', 'velocity', 'velocity_y', 'time_step']
    ST = ['position', 'steering_angle', 'velocity', 'orientation', 'yaw_rate', 'slip_angle', 'time_step']
    KS = ['position', 'steering_angle', 'velocity', 'orientation', 'time_step']
    MB = ['position', 'steering_angle', 'velocity', 'orientation', 'yaw_rate', 'roll_angle', 'roll_rate', 'pitch_angle',
          'pitch_rate', 'velocity_y', 'position_z', 'velocity_z', 'roll_angle_front', 'roll_rate_front',
          'velocity_y_front', 'position_z_front', 'velocity_z_front', 'roll_angle_rear', 'roll_rate_rear',
          'velocity_y_rear', 'position_z_rear', 'velocity_z_rear', 'left_front_wheel_angular_speed',
          'right_front_wheel_angular_speed', 'left_rear_wheel_angular_speed', 'right_rear_wheel_angular_speed',
          'delta_y_f', 'delta_y_r', 'time_step']
    Input = ['steering_angle_speed', 'acceleration', 'time_step']
    PMInput = ['acceleration', 'acceleration_y', 'time_step']


@unique
class XMLStateFields(Enum):
    """
    XML names of the state fields for vehicle models for different trajectory types.

    PM | ST | KS | MB -> Corresponding xml names of the state fields for trajectory states
    Input             -> XML names of the input fields for ST, KS, and MB vehicle models
    PMInput           -> XML names of the input fields for PM vehicle model.

    Note: If you change the order of xml names, don't forget to change the order on the StateFields enum as well,
    because the indexes have to match.
    """
    PM = [('x', 'y'), 'xVelocity', 'yVelocity', 'time']
    ST = [('x', 'y'), 'steeringAngle', 'velocity', 'orientation', 'yawRate', 'slipAngle', 'time']
    KS = [('x', 'y'), 'steeringAngle', 'velocity', 'orientation', 'time']
    MB = [('x', 'y'), 'steeringAngle', 'velocity', 'orientation', 'yawRate', 'rollAngle', 'rollRate', 'pitchAngle',
          'pitchRate', 'yVelocity', 'zPosition', 'zVelocity', 'rollAngleFront', 'rollRateFront',
          'yVelocityFront', 'zPositionFront', 'zVelocityFront', 'rollAngleRear', 'rollRateRear',
          'yVelocityRear', 'zPositionRear', 'zVelocityRear', 'leftFrontWheelAngularSpeed',
          'rightFrontWheelAngularSpeed', 'leftRearWheelAngularSpeed', 'rightRearWheelAngularSpeed',
          'deltaYf', 'deltaYr', 'time']
    Input = ['steeringAngleSpeed', 'acceleration', 'time']
    PMInput = ['xAcceleration', 'yAcceleration', 'time']


@unique
class StateType(Enum):
    """
    State Type enum class.

    PM | ST | KS | MB -> Corresponding state type for trajectory states
    Input             -> Input type for ST, KS, and MB vehicle models
    PMInput           -> Input type for PM vehicle model.
    """
    MB = 'mbState'
    ST = 'stState'
    KS = 'ksState'
    PM = 'pmState'
    Input = 'input'
    PMInput = 'pmInput'

    @property
    def fields(self) -> List[str]:
        """
        Returns the state fields for the state type.

        :return: State fields as list
        """
        return StateFields[self.name].value

    @property
    def xml_fields(self) -> List[str]:
        """
        Returns the xml state fields for the state type.

        :return: XML names of the state fields as list
        """
        return XMLStateFields[self.name].value

    @classmethod
    def get_state_type(cls, state: State, desired_vehicle_model: VehicleModel = None) -> 'StateType':
        """
        Returns the corresponding StateType for the given State object by matching State object's attributes
        to the state fields.

        :param state: CommonRoad State object
        :param desired_vehicle_model: check if given vehicle_model is supported first
        :return: corresponding StateType
        """
        # put desired_vehicle_model first
        attrs = state.attributes

        if desired_vehicle_model is not None:
            state_fields_all = [StateFields[str(desired_vehicle_model.name)], StateFields.Input, StateFields.PMInput]
            state_fields_add = []
            for sf in StateFields:
                if sf not in state_fields_all: state_fields_add.append(sf)

            state_fields_all += state_fields_add

            for state_fields in state_fields_all:
                if not len(attrs) >= len(state_fields.value): continue  # >=
                if not all([sf in attrs for sf in state_fields.value]): continue
                return cls[state_fields.name]
        else:
            state_fields_all = StateFields
            for state_fields in state_fields_all:
                if not len(attrs) == len(state_fields.value): continue  # ==
                if not all([sf in attrs for sf in state_fields.value]): continue
                return cls[state_fields.name]

        raise StateTypeException('Given state is not valid!')

    @classmethod
    def check_state_type(cls, vehicle_model: VehicleModel) -> bool:
        """
        Checks whether vehicle model can be supported by trajectory.
        :param vehicle_model: vehicle model enum
        :return: bool
        """
        StateFields(vehicle_model.name)


@unique
class TrajectoryType(Enum):
    """
    Trajectory Type enum class.

    PM | ST | KS | MB -> Corresponding trajectory type for the vehicle models
    Input             -> InputVector type for ST, KS, and MB vehicle models
    PMInput           -> InputVector type for PM vehicle model.
    """
    MB = 'mbTrajectory'
    ST = 'stTrajectory'
    KS = 'ksTrajectory'
    PM = 'pmTrajectory'
    Input = 'inputVector'
    PMInput = 'pmInputVector'

    @property
    def state_type(self) -> StateType:
        """
        Returns the StateType corresponding to the TrajectoryType

        :return: StateType
        """
        return StateType[self.name]

    @classmethod
    def get_trajectory_type(cls, trajectory: Trajectory,
                            desired_vehicle_model: VehicleModel = None) -> 'TrajectoryType':
        """
        Returns the corresponding TrajectoryType for the given Trajectory object based on the StateType of its states.

        :param trajectory: CommonRoad Trajectory object
        :param desired_vehicle_model: check if given vehicle_model is supported first
        :return: corresponding TrajectoryType
        """
        state_type = StateType.get_state_type(trajectory.state_list[0], desired_vehicle_model)
        return cls[state_type.name]

    def valid_vehicle_model(self, vehicle_model: VehicleModel) -> bool:
        """
        Checks whether given vehicle model is valid for the TrajectoryType.

        :param vehicle_model: CommonRoad enum for vehicle models
        :return: True if the vehicle model is valid for the TrajectoryType
        """
        return any([
            self.name == 'Input' and vehicle_model in [VehicleModel.KS, VehicleModel.ST, VehicleModel.MB],
            self.name == 'PMInput' and vehicle_model == VehicleModel.PM,
            self.name == vehicle_model.name
        ])


class SupportedCostFunctions(Enum):
    """
    Enum class for specifying which cost functions are supported for which vehicle model
    """
    PM = [CostFunction.JB1, CostFunction.WX1, CostFunction.MW1]
    ST = [cost_function for cost_function in CostFunction]  # Supports all cost functions
    KS = [cost_function for cost_function in CostFunction]  # Supports all cost functions
    MB = [cost_function for cost_function in CostFunction]  # Supports all cost functions


class PlanningProblemSolution:
    def __init__(self,
                 planning_problem_id: int,
                 vehicle_model: VehicleModel,
                 vehicle_type: VehicleType,
                 cost_function: CostFunction,
                 trajectory: Trajectory):
        """
        Constructor for the PlanningProblemSolution class.

        :param planning_problem_id: ID of the planning problem
        :param vehicle_model: VehicleModel used for the solution
        :param vehicle_type: VehicleType used for the solution
        :param cost_function: CostFunction the solution will be evaluated with
        :param trajectory: Ego vehicle's trajectory for the solution.
        """

        self.planning_problem_id = planning_problem_id
        self._vehicle_model = vehicle_model
        self.vehicle_type = vehicle_type
        self._cost_function = cost_function
        self._trajectory = trajectory
        self._trajectory_type = TrajectoryType.get_trajectory_type(self._trajectory, self.vehicle_model)

        self._check_trajectory_supported(self._vehicle_model, self._trajectory_type)
        self._check_cost_supported(self._vehicle_model, self._cost_function)

    @staticmethod
    def _check_cost_supported(vehicle_model: VehicleModel, cost_function: CostFunction) -> bool:
        """
        Checks whether given cost function is supported by the given vehicle model.

        :param vehicle_model: VehicleModel
        :param cost_function: CostFunction
        :return: True if supported.
        """
        supported_costs = SupportedCostFunctions[vehicle_model.name].value
        if cost_function not in supported_costs:
            raise SolutionException("Cost function %s isn't supported for %s model!" % (cost_function.name,
                                                                                        vehicle_model.name))
        return True

    def _check_trajectory_supported(self, vehicle_model: VehicleModel, trajectory_type: TrajectoryType) -> bool:
        """
        Checks whether given vehicle model is valid for the given trajectory type.

        :param vehicle_model: VehicleModel
        :param trajectory_type: TrajectoryType
        :return: True if valid.
        """
        if self._vehicle_model == VehicleModel.PM and self._trajectory_type == TrajectoryType.PM:
            for state in self._trajectory.state_list:
                if not hasattr(state, 'orientation'):
                    state.orientation = math.atan2(state.velocity_y, state.velocity)

        if not trajectory_type.valid_vehicle_model(vehicle_model):
            raise SolutionException('Vehicle model %s is not valid for the trajectory type %s!'
                                    % (vehicle_model.name, trajectory_type.name))
        return True

    @property
    def vehicle_model(self) -> VehicleModel:
        """ VehicleModel of the PlanningProblemSolution """
        return self._vehicle_model

    @vehicle_model.setter
    def vehicle_model(self, vehicle_model: VehicleModel):
        self._check_trajectory_supported(vehicle_model, self._trajectory_type)
        self._check_cost_supported(vehicle_model, self.cost_function)

        self._vehicle_model = vehicle_model

    @property
    def cost_function(self) -> CostFunction:
        """ CostFunction of the PlanningProblemSolution """
        return self._cost_function

    @cost_function.setter
    def cost_function(self, cost_function: CostFunction):
        self._check_cost_supported(self.vehicle_model, cost_function)
        self._cost_function = cost_function

    @property
    def trajectory(self) -> Trajectory:
        """ Trajectory of the PlanningProblemSolution """
        return self._trajectory

    @trajectory.setter
    def trajectory(self, trajectory: Trajectory):
        trajectory_type = TrajectoryType.get_trajectory_type(trajectory)
        self._check_trajectory_supported(self.vehicle_model, trajectory_type)

        self._trajectory = trajectory
        self._trajectory_type = trajectory_type

    @property
    def trajectory_type(self) -> TrajectoryType:
        """
        TrajectoryType of the PlanningProblemSolution.
        Dynamically assigned when there is a change of trajectory.
        """
        return self._trajectory_type

    @property
    def vehicle_id(self) -> str:
        """
        Returns the vehicle id as string.

        Example:
            VehicleModel = PM
            VehicleType = FORD_ESCORT
            Vehicle ID = PM1

        :return: Vehicle model ID
        """
        return self.vehicle_model.name + str(self.vehicle_type.value)

    @property
    def cost_id(self) -> str:
        """
        Returns cost function id as str.

        Example:
            CostFunction = JB1
            Cost ID = JB1

        :return: Cost function ID
        """
        return self.cost_function.name


class Solution:
    """Stores a solution to a CommonRoad benchmark and additional meta data."""

    def __init__(self,
                 scenario_id: ScenarioID,
                 planning_problem_solutions: List[PlanningProblemSolution],
                 date: datetime = datetime.today(),
                 computation_time: Union[float, None] = None,
                 processor_name: Union[str, None] = None):
        """
        :param scenario_id: Scenario ID of the Solution
        :param planning_problem_solutions: List of PlanningProblemSolution for corresponding
            to the planning problems of the scenario
        :param date: The date solution was produced. Default=datetime.today()
        :param computation_time: The computation time measured in seconds for the Solution. Default=None
        :param processor_name: The processor model used for the Solution. Determined automatically if set to 'auto'.
            Default=None.
        """
        self.scenario_id = scenario_id
        self._planning_problem_solutions: Dict[int, PlanningProblemSolution] = {}
        self.planning_problem_solutions = planning_problem_solutions
        self.date = date
        self._computation_time = None
        self.computation_time = computation_time
        self.processor_name = processor_name

    @property
    def planning_problem_solutions(self) -> List[PlanningProblemSolution]:
        return list(self._planning_problem_solutions.values())

    @planning_problem_solutions.setter
    def planning_problem_solutions(self, planning_problem_solutions: List[PlanningProblemSolution]):
        self._planning_problem_solutions = {s.planning_problem_id: s for s in planning_problem_solutions}

    @property
    def benchmark_id(self) -> str:
        """
        Returns the benchmark id of the solution as string.

        Example:
            Scenario ID = TEST
            VehicleModel = PM
            VehicleType = FORD_ESCORT
            CostFunction = JB1
            Version = 2020a

            Benchmark ID = PM1:JB1:TEST:2020a

        Collaborative Solution Example:
            Scenario ID = TEST
            1st VehicleModel = PM
            1st VehicleType = FORD_ESCORT
            1st CostFunction = JB1
            2nd VehicleModel = PM
            2nd VehicleType = VW_VANAGON
            2nd CostFunction = SA1
            Version = 2020a

            Benchmark ID = [PM1,PM3]:[JB1,SA1]:TEST:2020a

        :return: Benchmark ID
        """
        vehicle_ids = self.vehicle_ids
        cost_ids = self.cost_ids
        vehicles_str = vehicle_ids[0] if len(vehicle_ids) == 1 else '[%s]' % ','.join(vehicle_ids)
        costs_str = cost_ids[0] if len(cost_ids) == 1 else '[%s]' % ','.join(cost_ids)
        return '%s:%s:%s:%s' % (vehicles_str, costs_str, str(self.scenario_id), self.scenario_id.scenario_version)

    @property
    def vehicle_ids(self) -> List[str]:
        """
        Returns the list of vehicle ids of all PlanningProblemSolutions of the Solution

        Example:
            1st PlanningProblemSolution Vehicle ID = PM1
            2nd PlanningProblemSolution Vehicle ID = PM3

            Vehicle IDS = [PM1, PM3]

        :return: List of vehicle IDs
        """
        return [pp_solution.vehicle_id for pp_solution in self.planning_problem_solutions]

    @property
    def cost_ids(self) -> List[str]:
        """
        Returns the list of cost ids of all PlanningProblemSolutions of the Solution

        Example:
            1st PlanningProblemSolution Cost ID = JB1
            2nd PlanningProblemSolution Cost ID = SA1

            Cost IDS = [JB1, SA1]

        :return: List of cost function IDs
        """
        return [pp_solution.cost_id for pp_solution in self.planning_problem_solutions]

    @property
    def planning_problem_ids(self) -> List[int]:
        """
        Returns the list of planning problem ids of all PlanningProblemSolutions of the Solution

        Example:
            1st PlanningProblemSolution planning_problem_id = 0
            2nd PlanningProblemSolution planning_problem_id = 1

            planning_problem_ids = [0, 1]

        :return: List of planning problem ids
        """
        return [pp_solution.planning_problem_id for pp_solution in self.planning_problem_solutions]

    @property
    def trajectory_types(self) -> List[TrajectoryType]:
        """
        Returns the list of trajectory types of all PlanningProblemSolutions of the Solution

        Example:
            1st PlanningProblemSolution trajectory_type = TrajectoryType.PM
            2nd PlanningProblemSolution trajectory_type = TrajectoryType.KS

            trajectory_types = [TrajectoryType.PM, TrajectoryType.KS]

        :return: List of trajectory types
        """
        return [pp_solution.trajectory_type for pp_solution in self.planning_problem_solutions]

    @property
    def computation_time(self) -> Union[None, float]:
        """
        Return the computation time [s] for the trajectory.
        :return:
        """
        return self._computation_time

    @computation_time.setter
    def computation_time(self, computation_time):
        if computation_time is not None:
            assert is_real_number(computation_time), "<Solution> computation_time provided as type {}," \
                                                     "but expected type float," \
                                                     "measured in seconds!".format(type(computation_time))
            assert is_positive(computation_time), "<Solution> computation_time needs to be positive!"\
                .format(type(computation_time))
        self._computation_time = computation_time

    def create_dynamic_obstacle(self) -> Dict[int, DynamicObstacle]:
        """
        Creates dynamic obstacle(s) from solution(s) for every planning problem.
        :return:
        """
        obs = {}
        for pp_id, solution in self._planning_problem_solutions.items():
            shape = Rectangle(length=vehicle_parameters[solution.vehicle_type].l,
                              width=vehicle_parameters[solution.vehicle_type].w)
            trajectory = Trajectory(initial_time_step=solution.trajectory.initial_time_step + 1,
                              state_list=solution.trajectory.state_list[1:])
            prediction = TrajectoryPrediction(trajectory, shape=shape)
            obs[pp_id] = DynamicObstacle(obstacle_id=pp_id,
                                         obstacle_type=ObstacleType.CAR,
                                         obstacle_shape=shape,
                                         initial_state=solution.trajectory.state_list[0],
                                         prediction=prediction)

        return obs


class CommonRoadSolutionReader:
    """Reads solution xml files created with the CommonRoadSolutionWriter"""

    # TODO Add parser check

    @classmethod
    def open(cls, filepath: str) -> Solution:
        """
        Opens and parses the Solution XML file located on the given path.

        :param filepath: Path to the file.
        :return: Solution
        """
        tree = et.parse(filepath)
        root_node = tree.getroot()
        return cls._parse_solution(root_node)

    @classmethod
    def fromstring(cls, file: str) -> Solution:
        """
        Parses the given Solution XML string.

        :param file: xml file as string
        :return: Solution
        """
        root_node = et.fromstring(file)
        return cls._parse_solution(root_node)

    @classmethod
    def _parse_solution(cls, root_node: et.Element) -> Solution:
        """ Parses the Solution XML root node. """  # TODO
        benchmark_id, date, computation_time, processor_name = cls._parse_header(root_node)
        vehicle_ids, cost_ids, scenario_id = cls._parse_benchmark_id(benchmark_id)
        pp_solutions = [cls._parse_planning_problem_solution(vehicle_ids[idx], cost_ids[idx], trajectory_node)
                        for idx, trajectory_node in enumerate(root_node)]
        return Solution(scenario_id, pp_solutions, date, computation_time, processor_name)

    @staticmethod
    def _parse_header(root_node: et.Element) -> Tuple[str, Union[None, datetime], Union[None, float], Union[None, str]]:
        """ Parses the header attributes for the given Solution XML root node. """
        benchmark_id = root_node.get('benchmark_id', None)
        if not benchmark_id:
            SolutionException("Solution xml does not have a benchmark id!")

        date = root_node.attrib.get('date', None)  # None if not found
        if date is not None:
            date = datetime.strptime(date, '%Y-%m-%d')

        computation_time = root_node.attrib.get('computation_time', None)
        if computation_time is not None:
            computation_time = float(computation_time)

        processor_name = root_node.attrib.get('processor_name', None)

        return benchmark_id, date, computation_time, processor_name

    @classmethod
    def _parse_planning_problem_solution(cls, vehicle_id: str, cost_id: str,
                                         trajectory_node: et.Element) -> PlanningProblemSolution:
        """ Parses PlanningProblemSolution from the given XML node. """
        vehicle_model, vehicle_type = cls._parse_vehicle_id(vehicle_id)

        if cost_id not in [cfunc.name for cfunc in CostFunction]:
            raise SolutionReaderException("Invalid Cost ID: " + cost_id)
        cost_function = CostFunction[cost_id]

        pp_id, trajectory = cls._parse_trajectory(trajectory_node)
        return PlanningProblemSolution(pp_id, vehicle_model, vehicle_type, cost_function, trajectory)

    @classmethod
    def _parse_trajectory(cls, trajectory_node: et.Element) -> Tuple[int, Trajectory]:
        """ Parses Trajectory and planning problem id from the given XML node. """

        if trajectory_node.tag not in [ttype.value for ttype in TrajectoryType]:
            raise SolutionReaderException("Invalid Trajectory Type: " + trajectory_node.tag)
        trajectory_type = TrajectoryType(trajectory_node.tag)

        planning_problem_id = int(trajectory_node.get('planningProblem'))
        state_list = [cls._parse_state(trajectory_type.state_type, state_node) for state_node in trajectory_node]
        state_list = sorted(state_list, key=lambda state: state.time_step)
        return planning_problem_id, Trajectory(initial_time_step=state_list[0].time_step, state_list=state_list)

    @classmethod
    def _parse_sub_element(cls, state_node: et.Element, name: str, as_float: bool = True) -> Union[float, int]:
        """ Parses the sub elements from the given XML node. """
        elem = state_node.find(name)
        if elem is None:
            raise SolutionReaderException("Element '%s' couldn't be found in the xml node!" % name)
        value = float(elem.text) if as_float else int(elem.text)
        return value

    @classmethod
    def _parse_state(cls, state_type: StateType, state_node: et.Element) -> State:
        """ Parses State from the given XML node. """
        if not state_node.tag == state_type.value:
            raise SolutionReaderException("Given xml node is not a '%s' node!" % state_type.value)

        state_vals = {}
        for mapping in list(zip(state_type.xml_fields, state_type.fields)):
            xml_name = mapping[0]
            field_name = mapping[1]
            if isinstance(xml_name, tuple):
                state_vals[field_name] = np.array([cls._parse_sub_element(state_node, name) for name in xml_name])
            else:
                state_vals[field_name] = cls._parse_sub_element(state_node, xml_name, as_float=(not xml_name == 'time'))

        return State(**state_vals)

    @staticmethod
    def _parse_benchmark_id(benchmark_id: str) -> (List[str], List[str], str):
        """ Parses the given benchmark id string. """
        segments = benchmark_id.replace(' ', '').split(':')

        if len(segments) != 4:
            raise SolutionReaderException("Invalid Benchmark ID: " + benchmark_id)

        vehicle_model_ids = re.sub(r'[\[\]]', '', segments[0]).split(',')
        cost_function_ids = re.sub(r'[\[\]]', '', segments[1]).split(',')
        scenario_id = ScenarioID.from_benchmark_id(segments[2], segments[3])

        return vehicle_model_ids, cost_function_ids, scenario_id

    @staticmethod
    def _parse_vehicle_id(vehicle_id: str) -> Tuple[VehicleModel, VehicleType]:
        """ Parses the given vehicle id string. """
        if not len(vehicle_id) == 3:
            raise SolutionReaderException("Invalid Vehicle ID: " + vehicle_id)

        if not vehicle_id[:2] in [vmodel.name for vmodel in VehicleModel]:
            raise SolutionReaderException("Invalid Vehicle ID: " + vehicle_id)

        if not int(vehicle_id[2]) in [vtype.value for vtype in VehicleType]:
            raise SolutionReaderException("Invalid Vehicle ID: " + vehicle_id)

        return VehicleModel[vehicle_id[:2]], VehicleType(int(vehicle_id[2]))


class CommonRoadSolutionWriter:

    def __init__(self, solution: Solution):
        """
        Creates the xml file for the given solution that can be dumped as string, or written to file later on.

        :param solution: Solution.
        """
        assert isinstance(solution, Solution)
        self.solution = solution
        self._solution_root = self._serialize_solution(self.solution)

    @staticmethod
    def _get_processor_name() -> Union[str, None]:
        # TODO: compare cpu names with the list of cpu names used on web server

        delete_from_cpu_name = ['(R)', '(TM)']

        def strip_substrings(string: str):
            for del_str in delete_from_cpu_name:
                string = string.replace(del_str, '')
            return string

        if platform.system() == "Windows":
            name_tmp = platform.processor()
            for del_str in delete_from_cpu_name:
                name_tmp.replace(del_str, '')
            return strip_substrings(name_tmp)
        elif platform.system() == "Darwin":
            os.environ['PATH'] = os.environ['PATH'] + os.pathsep + '/usr/sbin'
            command = "sysctl -n machdep.cpu.brand_string"
            return str(subprocess.check_output(command, shell=True).strip())
        elif platform.system() == "Linux":
            command = "cat /proc/cpuinfo"
            all_info = str(subprocess.check_output(command, shell=True).strip())
            for line in all_info.split("\\n"):
                if "model name" in line:
                    name_tmp = re.sub(".*model name.*: ", "", line, 1)
                    return strip_substrings(name_tmp)
        return None

    @classmethod
    def _serialize_solution(cls, solution: Solution) -> et.Element:
        """ Serializes the given solution. """
        root_node = cls._create_root_node(solution)
        for pp_solution in solution.planning_problem_solutions:
            trajectory_node = cls._create_trajectory_node(pp_solution.trajectory_type,
                                                          pp_solution.planning_problem_id,
                                                          pp_solution.trajectory)
            root_node.append(trajectory_node)
        return root_node

    @classmethod
    def _create_root_node(cls, solution: Solution) -> et.Element:
        """ Creates the root node of the Solution XML. """
        root_node = et.Element('CommonRoadSolution')
        root_node.set('benchmark_id', solution.benchmark_id)
        if solution.date is not None: root_node.set('date', solution.date.strftime('%Y-%m-%d'))
        if solution.computation_time is not None: root_node.set('computation_time', str(solution.computation_time))
        processor_name = cls._get_processor_name() if solution.processor_name == 'auto' else solution.processor_name
        if processor_name is not None: root_node.set('processor_name', processor_name)
        return root_node

    @classmethod
    def _create_trajectory_node(cls, trajectory_type: TrajectoryType, pp_id: int, trajectory: Trajectory) -> et.Element:
        """ Creates the Trajectory XML Node for the given trajectory. """
        trajectory_node = et.Element(trajectory_type.value)
        trajectory_node.set('planningProblem', str(pp_id))
        for state in trajectory.state_list:
            state_node = cls._create_state_node(trajectory_type.state_type, state)
            trajectory_node.append(state_node)
        return trajectory_node

    @classmethod
    def _create_sub_element(cls, name: str, value: Union[float, int]) -> et.Element:
        """ Creates an XML element for the given value. """
        element = et.Element(name)
        element.text = str(np.float64(value) if isinstance(value, float) else value)
        return element

    @classmethod
    def _create_state_node(cls, state_type: StateType, state: State) -> et.Element:
        """ Creates XML nodes for the States of the Trajectory. """
        state_node = et.Element(state_type.value)
        for mapping in list(zip(state_type.xml_fields, state_type.fields)):
            xml_name = mapping[0]
            state_val = getattr(state, mapping[1])
            if isinstance(xml_name, tuple):
                for idx, name in enumerate(xml_name):
                    state_node.append(cls._create_sub_element(name, state_val[idx]))
            else:
                state_node.append(cls._create_sub_element(xml_name, state_val))
        return state_node

    def dump(self, pretty: bool = True) -> str:
        """
        Dumps the Solution XML as string.

        :param pretty: If set to true, prettifies the xml string.
        :return: string - Solution XML as string.
        """
        rough_string = et.tostring(self._solution_root, encoding='utf-8')

        if not pretty:
            return rough_string

        parsed = minidom.parseString(rough_string)
        return parsed.toprettyxml(indent="  ")

    def write_to_file(self, output_path: str = './', filename: str = None,
                      overwrite: bool = False, pretty: bool = True):
        """
        Writes the Solution XML to a file.

        :param output_path: Output dir where the Solution XML file should be written to. \
            Writes to the same folder where it is called from if not specified.
        :param filename: Name of the Solution XML file. If not specified, sets the name as 'solution_BENCHMARKID.xml' \
            where the BENCHMARKID is the benchmark_id of the solution.
        :param overwrite: If set to True, overwrites the file if it already exists.
        :param pretty: If set to True, prettifies the Solution XML string before writing to file.
        """
        filename = filename if filename is not None else 'solution_%s.xml' % self.solution.benchmark_id
        fullpath = os.path.join(output_path, filename) if filename is not None else os.path.join(output_path, filename)

        if not os.path.exists(os.path.dirname(fullpath)):
            raise NotADirectoryError("Directory %s does not exist." % os.path.dirname(fullpath))

        if os.path.exists(fullpath) and not overwrite:
            raise FileExistsError("File %s already exists. If you want to overwrite it set overwrite=True." % fullpath)

        with open(fullpath, 'w') as f:
            f.write(self.dump(pretty))



import math
import os
from collections import defaultdict
from typing import Dict, Set

import matplotlib as mpl
import matplotlib.artist as artists
import matplotlib.collections as collections
import matplotlib.colors
import matplotlib.patches as patches
import matplotlib.pyplot as plt
import matplotlib.text as text
import numpy as np
from matplotlib.animation import FuncAnimation
from matplotlib.colors import hsv_to_rgb, rgb_to_hsv, to_rgb, to_hex
from matplotlib.path import Path

import commonroad.geometry.shape
import commonroad.prediction.prediction
import commonroad.scenario.obstacle
from commonroad.common.util import Interval
from commonroad.geometry.shape import *
from commonroad.planning.goal import GoalRegion
from commonroad.planning.planning_problem import PlanningProblemSet, PlanningProblem
from commonroad.prediction.prediction import Occupancy, TrajectoryPrediction
from commonroad.scenario.lanelet import LaneletNetwork, LineMarking
from commonroad.scenario.obstacle import DynamicObstacle, StaticObstacle, ObstacleRole, SignalState, PhantomObstacle, \
    EnvironmentObstacle, Obstacle
from commonroad.scenario.scenario import Scenario
from commonroad.scenario.traffic_sign import TrafficLightState, TrafficLight, TrafficSign
from commonroad.scenario.trajectory import Trajectory, State
from commonroad.visualization.icons import supported_icons, get_obstacle_icon_patch
from commonroad.visualization.param_server import ParamServer
from commonroad.visualization.traffic_sign_v2 import draw_traffic_light_signs
from commonroad.visualization.util import LineDataUnits, collect_center_line_colors, get_arrow_path_at, colormap_idx, \
    line_marking_to_linestyle, traffic_light_color_dict, get_tangent_angle, approximate_bounding_box_dyn_obstacles, \
    get_vehicle_direction_triangle

__author__ = "Luis Gressenbuch"
__copyright__ = "TUM Cyber-Physical Systems Group"
__credits__ = [""]
__version__ = "2021.2"
__maintainer__ = "Luis Gressenbuch"
__email__ = "commonroad@lists.lrz.de"
__status__ = "Released"

traffic_sign_path = os.path.join(os.path.dirname(__file__), 'traffic_signs/')


class ZOrders:
    # Map
    LANELET_POLY = 9.0
    INCOMING_POLY = 9.1
    CROSSING_POLY = 9.2
    CENTER_BOUND = 10.0
    LIGHT_STATE_OTHER = 10.0
    LIGHT_STATE_GREEN = 10.05
    DIRECTION_ARROW = 10.1
    SUCCESSORS = 11.0
    STOP_LINE = 11.0
    RIGHT_BOUND = 12.0
    LEFT_BOUND = 12.0
    # Obstacles
    OBSTACLES = 20.0
    CAR_PATCH = 20.0
    # Labels
    LANELET_LABEL = 30.2
    STATE = 100.0
    LABELS = 1000.0
    # Values added to base value from drawing parameters
    INDICATOR_ADD = 0.2
    BRAKING_ADD = 0.2
    HORN_ADD = 0.1
    BLUELIGHT_ADD = 0.1


class MPRenderer(IRenderer):

    def __init__(self, draw_params: Union[ParamServer, dict, None] = None,
                 plot_limits: Union[List[Union[int, float]], None] = None, ax: Union[mpl.axes.Axes, None] = None,
                 figsize: Union[None, Tuple[float, float]]=None, focus_obstacle: Union[None, Obstacle] = None):
        """
        Creates an renderer for matplotlib

        :param draw_params: Default drawing params, if not supplied, default values are used.
        :param plot_limits: plotting limits. If not supplied, using `ax.autoscale()`.
        :param ax: Axis to use. If not supplied, `pyplot.gca()` is used.
        :param figsize: size of the figure
        :param focus_obstacle: if provided, the plot_limits are centered around center of obstacle at time_begin
        """

        self._plot_limits = None
        if draw_params is None:
            self.draw_params = ParamServer()
        elif isinstance(draw_params, dict):
            self.draw_params = ParamServer(params=draw_params)
        else:
            self.draw_params = draw_params
        self.plot_limits = plot_limits
        if ax is None:
            self.ax = plt.gca()
        else:
            self.ax = ax
        self.f = self.ax.figure

        if figsize is not None:
            self.f.set_size_inches(*figsize)

        # Draw elements
        self.dynamic_artists = []
        self.dynamic_collections = []
        self.static_artists = []
        self.static_collections = []
        self.obstacle_patches = []
        self.traffic_sign_artists = []
        self.traffic_signs = []
        self.traffic_sign_call_stack = tuple()
        self.traffic_sign_draw_params = self.draw_params
        # labels of dynamic elements
        self.dynamic_labels = []

        # current center of focus obstacle
        self.plot_center = None
        self.callbacks = defaultdict(list)

    @property
    def plot_limits(self):
        return self._plot_limits

    @plot_limits.setter
    def plot_limits(self, val: List[Union[float, int, List[Union[float, int]]]]):
        if val is not None and isinstance(val[0], List):
            self._plot_limits = val[0] + val[1]
        elif isinstance(val, List) or val == "auto":
            self._plot_limits = val
        elif val is not None:
            raise ValueError(f"Invalid plot_limit: {val}")

    @property
    def plot_limits_focused(self):
        """
        :returns: plot limits centered around focus_obstacle_id defined in draw_params
        """
        if self._plot_limits is not None and (self._plot_limits == "auto" or self.plot_center is None):
            return self._plot_limits
        elif self.plot_center is not None:
            plot_limits_f = np.array(self.plot_limits, dtype=int)
            plot_limits_f[:2] += int(self.plot_center[0])
            plot_limits_f[2:] += int(self.plot_center[1])
            return plot_limits_f

    def add_callback(self, event, func):
        self.callbacks[event].append(func)

    def draw_list(self, drawable_list: List[IDrawable], draw_params: Union[ParamServer, dict, None] = None,
                  call_stack: Optional[Tuple[str, ...]] = tuple()) -> None:
        """
        Simple wrapper to draw a list of drawable objects

        :param drawable_list: Objects to draw
        :param draw_params: parameters for plotting given by a nested dict
            that recreates the structure of an object or a ParamServer object
        :param call_stack: tuple of string containing the call stack,
            which allows for differentiation of plotting styles depending on the call stack
        :return: None
        """
        draw_params = self._get_draw_params(draw_params)
        for elem in drawable_list:
            elem.draw(self, draw_params, call_stack)

    def _get_draw_params(self, draw_params: Union[ParamServer, dict, None]) -> ParamServer:
        if draw_params is None:
            draw_params = self.draw_params
        elif isinstance(draw_params, dict):
            draw_params = ParamServer(params=draw_params, default=self.draw_params._params)
        return draw_params

    def clear(self, keep_static_artists=False) -> None:
        """
        Clears the internal drawing buffer

        :return: None
        """
        self.plot_center = None
        self.obstacle_patches.clear()
        self.traffic_signs.clear()
        self.traffic_sign_call_stack = tuple()
        self.traffic_sign_draw_params = self.draw_params
        self.dynamic_artists.clear()
        self.dynamic_collections.clear()
        self.traffic_sign_artists.clear()
        if keep_static_artists is False:
            self.static_artists.clear()
            self.static_collections.clear()

    def remove_dynamic(self) -> None:
        """
        Remove the dynamic objects from their current axis

        :return: None
        """
        for art in self.dynamic_artists:
            art.remove()

        # text artists cannot be removed -> set invisble
        for t in self.dynamic_labels:
            t.set_visible(False)
        self.dynamic_labels.clear()

    def render_dynamic(self) -> List[artists.Artist]:
        """
        Only render dynamic objects from buffer

        :return: List of drawn object's artists
        """
        artist_list = []
        self.traffic_sign_artists = draw_traffic_light_signs(self.traffic_signs, self.traffic_sign_draw_params,
                                                             self.traffic_sign_call_stack, self)
        for art in self.dynamic_artists:
            self.ax.add_artist(art)
            artist_list.append(art)
        for art in self.traffic_sign_artists:
            self.ax.add_artist(art)
            artist_list.append(art)
        for art in self.dynamic_artists:
            self.ax.add_artist(art)
        for col in self.dynamic_collections:
            self.ax.add_collection(col)
            artist_list.append(col)
        for t in self.dynamic_labels:
            self.ax.add_artist(t)

        self.obstacle_patches.sort(key=lambda x: x.zorder)
        patch_col = mpl.collections.PatchCollection(self.obstacle_patches, match_original=True,
                                                    zorder=ZOrders.OBSTACLES)
        self.ax.add_collection(patch_col)
        artist_list.append(patch_col)
        self.dynamic_artists = artist_list
        self._connect_callbacks()
        return artist_list

    def render_static(self) -> List[artists.Artist]:
        """
        Only render static objects from buffer

        :return: List of drawn object's artists
        """
        for col in self.static_collections:
            self.ax.add_collection(col)
        for art in self.static_artists:
            self.ax.add_artist(art)

        self._connect_callbacks()
        return self.static_collections + self.static_artists

    def render(self, show: bool = False, filename: str = None, keep_static_artists=False) -> List[artists.Artist]:
        """
        Render all objects from buffer

        :param show: Show the resulting figure
        :param filename: If provided, saves the figure to the provided file
        :return: List of drawn object's artists
        """
        self.ax.cla()
        artists_list = self.render_static()
        artists_list.extend(self.render_dynamic())

        if self.plot_limits is None:
            self.ax.autoscale(True)
        else:
            self.ax.set_xlim(self.plot_limits_focused[:2])
            self.ax.set_ylim(self.plot_limits_focused[2:])
        self.ax.set_aspect('equal')
        if filename is not None:
            self.f.savefig(filename, bbox_inches='tight')
        if show:
            self.f.show()

        if self.draw_params.by_callstack(param_path="axis_visible", call_stack=()) is False:
            self.ax.axes.xaxis.set_visible(False)
            self.ax.axes.yaxis.set_visible(False)

        self.clear(keep_static_artists)
        return artists_list

    def _connect_callbacks(self):
        """
        Connects collected callbacks with ax object.
        :return:
        """
        for event, funcs in self.callbacks.items():
            for fun in funcs:
                self.ax.callbacks.connect(event, fun)

        self.ax_updated = False

    def create_video(self, obj_lists: List[IDrawable], file_path: str, delta_time_steps: int = 1, plotting_horizon=0,
                     draw_params: Union[dict, ParamServer, None] = None, fig_size: Union[list, None] = None, dt=100,
                     dpi=120) -> None:
        """
        Creates a video of one or multiple CommonRoad objects in mp4, gif,
        or avi format.

        :param obj_lists: list of objects to be plotted.
        :param file_path: filename of generated video (ends on .mp4/.gif/.avi, default mp4, when nothing is specified)
        :param delta_time_steps: plot every delta_time_steps time steps of scenario
        :param plotting_horizon: time steps of prediction plotted in each frame
        :param draw_params: parameters for plotting given by a nested dict
            that recreates the structure of an object or a ParamServer object
        :param fig_size: size of the video
        :param dt: time step between frames in ms
        :param dpi: resolution of the video
        :return: None
        """
        draw_params = self._get_draw_params(draw_params)
        time_begin = draw_params['time_begin']
        time_end = draw_params['time_end']
        assert time_begin < time_end, '<video/create_scenario_video> ' \
                                      'time_begin=%i needs to smaller than ' \
                                      'time_end=%i.' % (time_begin, time_end)

        if fig_size is None:
            fig_size = [15, 8]

        self.ax.clear()
        self.f.set_size_inches(*fig_size)
        self.ax.set_aspect('equal')

        def init_frame():
            draw_params.update({'time_begin': time_begin, 'time_end': time_begin + delta_time_steps})
            self.draw_list(obj_lists, draw_params=draw_params)
            self.render_static()
            artists = self.render_dynamic()
            if self.plot_limits is None:
                self.ax.autoscale()
            elif self.plot_limits == 'auto':
                limits = approximate_bounding_box_dyn_obstacles(obj_lists, time_begin)
                if limits is not None:
                    self.ax.xlim(limits[0][0] - 10, limits[0][1] + 10)
                    self.ax.ylim(limits[1][0] - 10, limits[1][1] + 10)
                else:
                    self.ax.autoscale()
            else:
                self.ax.set_xlim(self.plot_limits_focused[0], self.plot_limits_focused[1])
                self.ax.set_ylim(self.plot_limits_focused[2], self.plot_limits_focused[3])

            if draw_params.by_callstack(param_path="axis_visible", call_stack=()) is False:
                self.ax.axes.xaxis.set_visible(False)
                self.ax.axes.yaxis.set_visible(False)
            return artists

        def update(frame=0):
            draw_params.update({'time_begin': time_begin + delta_time_steps * frame,
                                'time_end': time_begin + min(frame_count, delta_time_steps * frame + plotting_horizon)})
            self.remove_dynamic()
            self.clear()
            self.draw_list(obj_lists, draw_params=draw_params)
            artists = self.render_dynamic()
            if self.plot_limits is None:
                self.ax.autoscale()
            elif self.plot_limits == 'auto':
                limits = approximate_bounding_box_dyn_obstacles(obj_lists, time_begin)
                if limits is not None:
                    self.ax.xlim(limits[0][0] - 10, limits[0][1] + 10)
                    self.ax.ylim(limits[1][0] - 10, limits[1][1] + 10)
                else:
                    self.ax.autoscale()
            else:
                self.ax.set_xlim(self.plot_limits_focused[0], self.plot_limits_focused[1])
                self.ax.set_ylim(self.plot_limits_focused[2], self.plot_limits_focused[3])
            return artists

        # Min frame rate is 1 fps
        dt = min(1000.0, dt)
        frame_count = (time_end - time_begin) // delta_time_steps
        plt.ioff()
        # Interval determines the duration of each frame in ms
        anim = FuncAnimation(self.f, update, frames=frame_count, init_func=init_frame, blit=False, interval=dt)

        if not any([file_path.endswith('.mp4'), file_path.endswith('.gif'), file_path.endswith('.avi')]):
            file_path += '.mp4'
        fps = int(math.ceil(1000.0 / dt))
        interval_seconds = dt / 1000.0
        anim.save(file_path, dpi=dpi, writer='ffmpeg', fps=fps,
                  extra_args=["-g", "1", "-keyint_min", str(interval_seconds)])
        self.clear()
        self.ax.clear()

    def add_legend(self, legend: Dict[Tuple[str, ...], str],
                   draw_params: Union[ParamServer, dict, None] = None) -> None:
        """
        Adds legend with color of objects specified by legend.keys() and
        texts specified by legend.values().

        :param legend: color of objects specified by path in legend.keys() and texts specified by legend.values()
        :param draw_params: draw parameters used for plotting (color is extracted using path in legend.keys())
        :return: None
        """
        draw_params = self._get_draw_params(draw_params)
        handles = []
        for obj_name, text in legend.items():
            try:
                color = draw_params[obj_name]
            except KeyError:
                color = None
            if color is not None:
                handles.append(mpl.patches.Patch(color=color, label=text))

        legend = self.ax.legend(handles=handles)
        legend.set_zorder(ZOrders.LABELS)

    def draw_scenario(self, obj: Scenario, draw_params: Union[ParamServer, dict, None],
                      call_stack: Tuple[str, ...]) -> None:
        """
        :param obj: object to be plotted
        :param draw_params: parameters for plotting given by a nested dict
            that recreates the structure of an object or a ParamServer object
        :param call_stack: tuple of string containing the call stack,
            which allows for differentiation of plotting styles depending on the call stack
        :return: None
        """
        draw_params = self._get_draw_params(draw_params)
        call_stack = tuple(list(call_stack) + ['scenario'])
        obj.lanelet_network.draw(self, draw_params, call_stack)

        # draw only obstacles inside plot limits
        focus_obstacle_id = draw_params.by_callstack(call_stack, ('focus_obstacle_id',))
        if focus_obstacle_id is False and type(self.plot_limits) == list:
            time_begin = draw_params.by_callstack(call_stack, ('time_begin',))
            # dynamic obstacles
            dyn_obs = obj.obstacles_by_position_intervals([Interval(self.plot_limits[0], self.plot_limits[1]),
                                                           Interval(self.plot_limits[2], self.plot_limits[3])],
                                                          tuple([ObstacleRole.DYNAMIC]), time_begin)

            # static obstacles
            static_obs = obj.obstacles_by_position_intervals([Interval(self.plot_limits[0], self.plot_limits[1]),
                                                              Interval(self.plot_limits[2], self.plot_limits[3])],
                                                             tuple([ObstacleRole.STATIC]))
        else:
            dyn_obs = obj.dynamic_obstacles
            static_obs = obj.static_obstacles
        # Draw all objects
        for o in dyn_obs:
            o.draw(self, draw_params, call_stack)
        for o in static_obs:
            o.draw(self, draw_params, call_stack)

    def draw_static_obstacle(self, obj: StaticObstacle, draw_params: Union[ParamServer, dict, None],
                             call_stack: Tuple[str, ...]) -> None:
        """
        :param obj: object to be plotted
        :param draw_params: parameters for plotting given by a nested dict
            that recreates the structure of an object or a ParamServer object
        :param call_stack: tuple of string containing the call stack,
            which allows for differentiation of plotting styles depending on the call stack
        :return: None
        """
        draw_params = self._get_draw_params(draw_params)
        time_begin = draw_params.by_callstack(tuple(), ('time_begin',))
        call_stack = tuple(list(call_stack) + ['static_obstacle'])
        occ = obj.occupancy_at_time(time_begin)
        self._draw_occupancy(occ, obj.initial_state, draw_params, call_stack)

    def _draw_occupancy(self, occ: Occupancy, state: State, draw_params: Union[ParamServer, dict, None],
                        call_stack: Tuple[str, ...]) -> None:
        if occ is not None:
            occ.draw(self, draw_params, call_stack)
        if state is not None and state.is_uncertain_position:
            zorder_polygon = draw_params.by_callstack(call_stack, ("occupancy", "shape", "polygon", "zorder")) + 0.1
            zorder_rectangle = draw_params.by_callstack(call_stack, ("occupancy", "shape", "rectancle", "zorder")) + 0.1
            zorder_circle = draw_params.by_callstack(call_stack, ("occupancy", "shape", "circle", "zorder")) + 0.1
            draw_params = {"occupancy": {"uncertain_position": {
                "shape": {"polygon": {"zorder": zorder_polygon}, "rectangle": {"zorder": zorder_rectangle},
                          "circle": {"zorder": zorder_circle}}}}}
            state.position.draw(self, draw_params, call_stack + ('occupancy', 'uncertain_position'))

    def draw_dynamic_obstacle(self, obj: DynamicObstacle, draw_params: Union[ParamServer, dict, None],
                              call_stack: Tuple[str, ...]) -> None:
        """
        :param obj: object to be plotted
        :param draw_params: parameters for plotting given by a nested dict
            that recreates the structure of an object or a ParamServer object
        :param call_stack: tuple of string containing the call stack,
            which allows for differentiation of plotting styles depending on the call stack
        :return: None
        """
        draw_params = self._get_draw_params(draw_params)

        time_begin = draw_params.by_callstack(call_stack, ('time_begin',))
        time_end = draw_params.by_callstack(call_stack, ('time_end',))
        focus_obstacle_id = draw_params.by_callstack(call_stack, ('focus_obstacle_id',))
        call_stack = tuple(list(call_stack) + ['dynamic_obstacle'])
        draw_icon = draw_params.by_callstack(call_stack, 'draw_icon')
        show_label = draw_params.by_callstack(call_stack, 'show_label')
        draw_shape = draw_params.by_callstack(call_stack, 'draw_shape')
        draw_direction = draw_params.by_callstack(call_stack, 'draw_direction')
        draw_initial_state = draw_params.by_callstack(call_stack, 'draw_initial_state')
        draw_occupancies = draw_params.by_callstack(call_stack, ('occupancy', 'draw_occupancies'))
        draw_signals = draw_params.by_callstack(call_stack, 'draw_signals')
        draw_trajectory = draw_params.by_callstack(call_stack, ('trajectory', 'draw_trajectory'))

        draw_history = draw_params.by_callstack(call_stack, ('history', 'draw_history'))

        if obj.prediction is None or \
            obj.prediction.final_time_step < time_begin or obj.initial_state.time_step > time_end:
            return

        if draw_history and isinstance(obj.prediction, commonroad.prediction.prediction.TrajectoryPrediction):
            self._draw_history(obj, call_stack, draw_params)

        # draw car icon
        if draw_icon and obj.obstacle_type in supported_icons() and type(
                obj.prediction) == commonroad.prediction.prediction.TrajectoryPrediction:

            try:
                length = obj.obstacle_shape.length
                width = obj.obstacle_shape.width
            except AttributeError:
                draw_shape = True
                draw_icon = False

            if draw_icon:
                if time_begin == 0:
                    inital_state = obj.initial_state
                else:
                    inital_state = obj.prediction.trajectory.state_at_time_step(time_begin)
                if inital_state is not None:
                    call_stack_tmp = call_stack + ('occupancy', 'shape', 'polygon')

                    facecolor = draw_params.by_callstack(call_stack_tmp, 'facecolor')
                    edgecolor = draw_params.by_callstack(call_stack_tmp, 'edgecolor')

                    self.obstacle_patches.extend(
                        get_obstacle_icon_patch(obj.obstacle_type, inital_state.position[0], inital_state.position[1],
                                                inital_state.orientation, vehicle_length=length, vehicle_width=width,
                                                vehicle_color=facecolor, edgecolor=edgecolor, zorder=ZOrders.CAR_PATCH))
        elif draw_icon is True:
            draw_shape = True

        # draw shape
        if draw_shape:
            if time_begin < obj.initial_state.time_step <= time_end:
                time_begin = obj.initial_state.time_step
            veh_occ = obj.occupancy_at_time(time_begin)
            self._draw_occupancy(veh_occ, obj.initial_state, draw_params, call_stack + ('vehicle_shape',))
            if draw_direction and veh_occ is not None and type(veh_occ.shape) == Rectangle:
                v_tri = get_vehicle_direction_triangle(veh_occ.shape)
                self.draw_polygon(v_tri, draw_params, call_stack + ('vehicle_shape', 'direction'))

            if draw_signals:
                sig = obj.signal_state_at_time_step(time_begin)
                occ = veh_occ
                if occ is not None and sig is not None:
                    self._draw_signal_state(sig, occ, draw_params, call_stack)

        # draw occupancies
        if (draw_occupancies == 1 or type(obj.prediction) == commonroad.prediction.prediction.SetBasedPrediction):
            if draw_shape:
                # occupancy already plotted
                time_begin_occ = time_begin + 1
            else:
                time_begin_occ = time_begin

            for time_step in range(time_begin_occ, time_end):
                state = None
                if isinstance(obj.prediction, TrajectoryPrediction):
                    state = obj.prediction.trajectory.state_at_time_step(time_step)
                occ = obj.occupancy_at_time(time_step)
                self._draw_occupancy(occ, state, draw_params, call_stack)

        # draw trajectory
        if draw_trajectory and type(obj.prediction) == commonroad.prediction.prediction.TrajectoryPrediction:
            obj.prediction.trajectory.draw(self, draw_params, call_stack)

        # get state
        state = None
        if time_begin == 0:
            state = obj.initial_state
        elif type(obj.prediction) == commonroad.prediction.prediction.TrajectoryPrediction:
            state = obj.prediction.trajectory.state_at_time_step(time_begin)

        # set plot center state
        if focus_obstacle_id == obj.obstacle_id and state is not None:
            self.plot_center = state.position

        # draw label
        if show_label:
            if state is not None:
                position = state.position
                self.dynamic_labels.append(
                        text.Text(position[0] + 0.5, position[1], str(obj.obstacle_id), clip_on=True,
                                  zorder=ZOrders.LABELS))

        # draw initial state
        if draw_initial_state:
            state.draw(self, draw_params, call_stack)

    def draw_phantom_obstacle(self, obj: PhantomObstacle, draw_params: Union[ParamServer, dict, None],
                              call_stack: Tuple[str, ...]) -> None:
        """
        :param obj: object to be plotted
        :param draw_params: parameters for plotting given by a nested dict
            that recreates the structure of an object or a ParamServer object
        :param call_stack: tuple of string containing the call stack,
            which allows for differentiation of plotting styles depending on the call stack
        :return: None
        """
        draw_params = self._get_draw_params(draw_params)

        time_begin = draw_params.by_callstack(call_stack, ('time_begin',))
        time_end = draw_params.by_callstack(call_stack, ('time_end',))

        call_stack = tuple(list(call_stack) + ['dynamic_obstacle'])
        draw_shape = draw_params.by_callstack(call_stack, 'draw_shape')
        draw_occupancies = draw_params.by_callstack(call_stack, ('occupancy', 'draw_occupancies'))

        # draw occupancies
        if (draw_occupancies == 1 or (
                draw_occupancies == 0 and type(obj.prediction) == commonroad.prediction.prediction.SetBasedPrediction)):
            if draw_shape:
                # occupancy already plotted
                time_begin_occ = time_begin + 1
            else:
                time_begin_occ = time_begin

            for time_step in range(time_begin_occ, time_end):
                occ = obj.occupancy_at_time(time_step)
                if occ is not None:
                    occ.draw(self, draw_params, call_stack)

        # draw shape
        if draw_shape:
            occ = obj.occupancy_at_time(time_begin)
            if occ is not None:
                occ.draw(self, draw_params, call_stack + ('vehicle_shape',))

    def draw_environment_obstacle(self, obj: EnvironmentObstacle, draw_params: Union[ParamServer, dict, None],
                                  call_stack: Tuple[str, ...]) -> None:
        """
        :param obj: object to be plotted
        :param draw_params: parameters for plotting given by a nested dict
            that recreates the structure of an object or a ParamServer object
        :param call_stack: tuple of string containing the call stack,
            which allows for differentiation of plotting styles depending on the call stack
        :return: None
        """
        draw_params = self._get_draw_params(draw_params)
        time_begin = draw_params.by_callstack(tuple(), ('time_begin',))
        call_stack = tuple(list(call_stack) + ['static_obstacle'])
        obj.occupancy_at_time(time_begin).draw(self, draw_params, call_stack)

    def _draw_history(self, dyn_obs: DynamicObstacle, call_stack: Tuple[str, ...], draw_params: ParamServer):
        """
        Draws history occupancies of the dynamic obstacle

        :param call_stack: tuple of string containing the call stack,
        which allows for differentiation of plotting styles
               depending on the call stack of drawing functions
        :param draw_params: parameters for plotting given by a nested dict
        that recreates the structure of an object or a ParamServer object
        :param dyn_obs: the dynamic obstacle
        :return:
        """
        time_begin = draw_params['time_begin']
        history_base_color = draw_params.by_callstack(call_stack,
                                                      ('vehicle_shape', 'occupancy', 'shape', 'rectangle', 'facecolor'))
        history_callstack = call_stack + ('history',)
        history_steps = draw_params.by_callstack(history_callstack, 'steps')
        history_fade_factor = draw_params.by_callstack(history_callstack, 'fade_factor')
        history_step_size = draw_params.by_callstack(history_callstack, 'step_size')
        history_base_color = rgb_to_hsv(to_rgb(history_base_color))
        for history_idx in range(history_steps, 0, -1):
            time_step = time_begin - history_idx * history_step_size
            occ = dyn_obs.occupancy_at_time(time_step)
            if occ is not None:
                color_hsv_new = history_base_color.copy()
                color_hsv_new[2] = max(0, color_hsv_new[2] - history_fade_factor * history_idx)
                color_hex_new = to_hex(hsv_to_rgb(color_hsv_new))
                draw_params[history_callstack + ('occupancy', 'shape', 'rectangle', 'facecolor')] = color_hex_new
                draw_params[history_callstack + ('occupancy', 'shape', 'circle', 'facecolor')] = color_hex_new
                draw_params[history_callstack + ('occupancy', 'shape', 'polygon', 'facecolor')] = color_hex_new
                occ.draw(self, draw_params, history_callstack)

    def draw_trajectory(self, obj: Trajectory, draw_params: Union[ParamServer, dict, None],
                        call_stack: Tuple[str, ...]) -> None:
        """
        :param obj: object to be plotted
        :param draw_params: parameters for plotting given by a nested dict
            that recreates the structure of an object or a ParamServer object
        :param call_stack: tuple of string containing the call stack,
            which allows for differentiation of plotting styles depending on the call stack
        :return: None
        """
        draw_params = self._get_draw_params(draw_params)

        time_begin = draw_params.by_callstack(call_stack, 'time_begin')
        time_end = draw_params.by_callstack(call_stack, 'time_end')

        call_stack = call_stack + ('trajectory',)
        line_color = draw_params.by_callstack(call_stack, 'facecolor')

        line_width = draw_params.by_callstack(call_stack, 'line_width')
        draw_continuous = draw_params.by_callstack(call_stack, 'draw_continuous')
        z_order = draw_params.by_callstack(call_stack, 'z_order')

        if time_begin == time_end:
            return

        traj_states = [obj.state_at_time_step(t) for t in range(time_begin, time_end) if
                       obj.state_at_time_step(t) is not None]
        position_sets = [s.position for s in traj_states if s.is_uncertain_position]
        traj_points = [s.position for s in traj_states if not s.is_uncertain_position]

        traj_points = np.array(traj_points)

        # Draw certain states
        if draw_continuous:
            path = mpl.path.Path(traj_points, closed=False)
            self.obstacle_patches.append(
                    mpl.patches.PathPatch(path, color=line_color, lw=line_width, zorder=z_order, fill=False))
        else:
            self.dynamic_collections.append(
                    collections.EllipseCollection(np.ones([traj_points.shape[0], 1]) * line_width,
                                                  np.ones([traj_points.shape[0], 1]) * line_width,
                                                  np.zeros([traj_points.shape[0], 1]), offsets=traj_points, units='xy',
                                                  linewidths=0, zorder=z_order, transOffset=self.ax.transData,
                                                  facecolor=line_color))

        # Draw uncertain states
        for p in position_sets:
            p.draw(self, draw_params, call_stack)

    def draw_trajectories(self, obj: List[Trajectory], draw_params: Union[ParamServer, dict, None],
                          call_stack: Tuple[str, ...]) -> None:
        draw_params = self._get_draw_params(draw_params)
        unique_colors = draw_params.by_callstack(call_stack, ('trajectory', 'unique_colors'))
        if unique_colors:
            cmap = colormap_idx(len(obj))
            for i, traj in enumerate(obj):
                draw_params['trajectory', 'facecolor'] = mpl.colors.to_hex(cmap(i))
                traj.draw(self, draw_params, call_stack)
        else:
            self.draw_list(obj, draw_params)

    def draw_polygon(self, vertices, draw_params: Union[ParamServer, dict, None], call_stack: Tuple[str, ...]) -> None:
        """
        Draws a polygon shape

        :param vertices: vertices of the polygon
        :param draw_params: parameters for plotting given by a nested dict
            that recreates the structure of an object or a ParamServer object
        :param call_stack: tuple of string containing the call stack,
            which allows for differentiation of plotting styles depending on the call stack
        :return: None
        """
        draw_params = self._get_draw_params(draw_params)
        call_stack = tuple(list(call_stack) + ['shape', 'polygon'])
        facecolor = draw_params.by_callstack(call_stack, 'facecolor')
        edgecolor = draw_params.by_callstack(call_stack, 'edgecolor')
        zorder = draw_params.by_callstack(call_stack, 'zorder')
        opacity = draw_params.by_callstack(call_stack, 'opacity')
        linewidth = draw_params.by_callstack(call_stack, 'linewidth')
        antialiased = draw_params.by_callstack(call_stack, 'antialiased')
        self.obstacle_patches.append(
                mpl.patches.Polygon(vertices, closed=True, facecolor=facecolor, edgecolor=edgecolor, zorder=zorder,
                                    alpha=opacity, linewidth=linewidth, antialiased=antialiased))

    def draw_rectangle(self, vertices: np.ndarray, draw_params: Union[ParamServer, dict, None],
                       call_stack: Tuple[str, ...]) -> None:
        """
        Draws a rectangle shape

        :param vertices: vertices of the rectangle
        :param draw_params: parameters for plotting given by a nested dict that
            recreates the structure of an object,
        :param call_stack: tuple of string containing the call stack,
            which allows for differentiation of plotting styles depending on the call stack
        :return: None
        """
        draw_params = self._get_draw_params(draw_params)
        call_stack = tuple(list(call_stack) + ['shape', 'rectangle'])

        facecolor = draw_params.by_callstack(call_stack, 'facecolor')
        edgecolor = draw_params.by_callstack(call_stack, 'edgecolor')
        zorder = draw_params.by_callstack(call_stack, 'zorder')
        opacity = draw_params.by_callstack(call_stack, 'opacity')
        linewidth = draw_params.by_callstack(call_stack, 'linewidth')
        antialiased = draw_params.by_callstack(call_stack, 'antialiased')

        self.obstacle_patches.append(
                mpl.patches.Polygon(vertices, closed=True, zorder=zorder, facecolor=facecolor, edgecolor=edgecolor,
                                    alpha=opacity, antialiased=antialiased, linewidth=linewidth))

    def draw_ellipse(self, center: List[float], radius_x: float, radius_y: float,
                     draw_params: Union[ParamServer, dict, None], call_stack: Tuple[str, ...]) -> None:
        """
        Draws a circle shape

        :param ellipse: center position of the ellipse
        :param radius_x: radius of the ellipse along the x-axis
        :param radius_y: radius of the ellipse along the y-axis
        :param draw_params: parameters for plotting given by a nested dict that
            recreates the structure of an object,
        :param call_stack: tuple of string containing the call stack,
            which allows for differentiation of plotting styles depending on the call stack
        :return: None
        """
        draw_params = self._get_draw_params(draw_params)
        call_stack = tuple(list(call_stack) + ['shape', 'circle'])
        facecolor = draw_params.by_callstack(call_stack, 'facecolor')
        edgecolor = draw_params.by_callstack(call_stack, 'edgecolor')
        zorder = draw_params.by_callstack(call_stack, 'zorder')
        opacity = draw_params.by_callstack(call_stack, 'opacity')
        linewidth = draw_params.by_callstack(call_stack, 'linewidth')

        self.obstacle_patches.append(
                mpl.patches.Ellipse(center, 2 * radius_x, 2 * radius_y, facecolor=facecolor, edgecolor=edgecolor,
                                    zorder=zorder, linewidth=linewidth, alpha=opacity))

    def draw_state(self, state: State, draw_params: Union[ParamServer, dict, None],
                   call_stack: Tuple[str, ...] = None) -> None:
        """
        Draws a state as an arrow of its velocity vector

        :param state: state to be plotted
        :param draw_params: parameters for plotting given by a nested dict
            that recreates the structure of an object or a ParamServer object
        :param call_stack: tuple of string containing the call stack,
            which allows for differentiation of plotting styles depending on the call stack
        :return: None
        """
        draw_params = self._get_draw_params(draw_params)
        call_stack += 'state',
        scale_factor = draw_params.by_callstack(call_stack, 'scale_factor')
        arrow_args = draw_params.by_callstack(call_stack, 'kwargs')
        draw_arrow = draw_params.by_callstack(call_stack, 'draw_arrow')
        if draw_arrow:
            cos = math.cos(state.orientation)
            sin = math.sin(state.orientation)
            x = state.position[0]
            y = state.position[1]
            self.obstacle_patches.append(mpl.patches.Arrow(x=x, y=y, dx=state.velocity * cos * scale_factor,
                                                           dy=state.velocity * sin * scale_factor, zorder=ZOrders.STATE,
                                                           **arrow_args))
        else:
            self.obstacle_patches.append(mpl.patches.Circle(state.position, radius=0.5, zorder=ZOrders.STATE))

    def draw_lanelet_network(self, obj: LaneletNetwork, draw_params: Union[ParamServer, dict, None],
                             call_stack: Tuple[str, ...]) -> None:
        """
        Draws a lanelet network

        :param obj: object to be plotted
        :param draw_params: parameters for plotting given by a nested dict that
            recreates the structure of an object,
        :param call_stack: tuple of string containing the call stack,
            which allows for differentiation of plotting styles depending on the call stack
        :return: None
        """
        draw_params = self._get_draw_params(draw_params)
        call_stack = tuple(list(call_stack) + ['lanelet_network'])

        traffic_lights = obj.traffic_lights
        traffic_signs = obj.traffic_signs
        intersections = obj.intersections
        lanelets = obj.lanelets

        time_begin = draw_params.by_callstack(call_stack, ('time_begin',))
        if traffic_lights is not None:
            draw_traffic_lights = draw_params.by_callstack(call_stack, ('traffic_light', 'draw_traffic_lights'))

            traffic_light_colors = draw_params.by_callstack(call_stack, ('traffic_light'))
        else:
            draw_traffic_lights = False

        if traffic_signs is not None:
            draw_traffic_signs = draw_params.by_callstack(call_stack, ('traffic_sign', 'draw_traffic_signs'))
            show_traffic_sign_label = draw_params.by_callstack(call_stack, ('traffic_sign', 'show_label'))
        else:
            draw_traffic_signs = show_traffic_sign_label = False

        if intersections is not None and len(intersections) > 0:
            draw_intersections = draw_params.by_callstack(call_stack, ('intersection', 'draw_intersections'))
        else:
            draw_intersections = False

        if draw_intersections is True:
            draw_incoming_lanelets = draw_params.by_callstack(call_stack, ('intersection', 'draw_incoming_lanelets'))
            incoming_lanelets_color = draw_params.by_callstack(call_stack, ('intersection', 'incoming_lanelets_color'))
            draw_crossings = draw_params.by_callstack(call_stack, ('intersection', 'draw_crossings'))
            crossings_color = draw_params.by_callstack(call_stack, ('intersection', 'crossings_color'))
            draw_successors = draw_params.by_callstack(call_stack, ('intersection', 'draw_successors'))
            successors_left_color = draw_params.by_callstack(call_stack, ('intersection', 'successors_left_color'))
            successors_straight_color = draw_params.by_callstack(call_stack,
                                                                 ('intersection', 'successors_straight_color'))
            successors_right_color = draw_params.by_callstack(call_stack, ('intersection', 'successors_right_color'))
            show_intersection_labels = draw_params.by_callstack(call_stack, ('intersection', 'show_label'))
        else:
            draw_incoming_lanelets = draw_crossings = draw_successors = show_intersection_labels = False

        left_bound_color = draw_params.by_callstack(call_stack, ('lanelet', 'left_bound_color'))
        right_bound_color = draw_params.by_callstack(call_stack, ('lanelet', 'right_bound_color'))
        center_bound_color = draw_params.by_callstack(call_stack, ('lanelet', 'center_bound_color'))
        unique_colors = draw_params.by_callstack(call_stack, ('lanelet', 'unique_colors'))
        draw_stop_line = draw_params.by_callstack(call_stack, ('lanelet', 'draw_stop_line'))
        stop_line_color = draw_params.by_callstack(call_stack, ('lanelet', 'stop_line_color'))
        draw_line_markings = draw_params.by_callstack(call_stack, ('lanelet', 'draw_line_markings'))
        show_label = draw_params.by_callstack(call_stack, ('lanelet', 'show_label'))
        draw_border_vertices = draw_params.by_callstack(call_stack, ('lanelet', 'draw_border_vertices'))
        draw_left_bound = draw_params.by_callstack(call_stack, ('lanelet', 'draw_left_bound'))
        draw_right_bound = draw_params.by_callstack(call_stack, ('lanelet', 'draw_right_bound'))
        draw_center_bound = draw_params.by_callstack(call_stack, ('lanelet', 'draw_center_bound'))
        draw_start_and_direction = draw_params.by_callstack(call_stack, ('lanelet', 'draw_start_and_direction'))
        draw_linewidth = draw_params.by_callstack(call_stack, ('lanelet', 'draw_linewidth'))
        fill_lanelet = draw_params.by_callstack(call_stack, ('lanelet', 'fill_lanelet'))
        facecolor = draw_params.by_callstack(call_stack, ('lanelet', 'facecolor'))
        antialiased = draw_params.by_callstack(call_stack, 'antialiased')

        draw_lanlet_ids = draw_params.by_callstack(call_stack, 'draw_ids')

        colormap_tangent = draw_params.by_callstack(call_stack, ('lanelet', 'colormap_tangent'))

        # Collect lanelets
        incoming_lanelets = set()
        incomings_left = {}
        incomings_id = {}
        crossings = set()
        all_successors = set()
        successors_left = set()
        successors_straight = set()
        successors_right = set()
        if draw_intersections:
            # collect incoming lanelets
            if draw_incoming_lanelets:
                incomings: List[set] = []
                inc_2_intersections = obj.map_inc_lanelets_to_intersections
                for intersection in intersections:
                    for incoming in intersection.incomings:
                        incomings.append(incoming.incoming_lanelets)
                        for l_id in incoming.incoming_lanelets:
                            incomings_left[l_id] = incoming.left_of
                            incomings_id[l_id] = incoming.incoming_id
                incoming_lanelets: Set[int] = set.union(*incomings)

            if draw_crossings:
                tmp_list: List[set] = [intersection.crossings for intersection in intersections]
                crossings: Set[int] = set.union(*tmp_list)

            if draw_successors:
                tmp_list: List[set] = [incoming.successors_left for intersection in intersections for incoming in
                                       intersection.incomings]
                successors_left: Set[int] = set.union(*tmp_list)
                tmp_list: List[set] = [incoming.successors_straight for intersection in intersections for incoming in
                                       intersection.incomings]
                successors_straight: Set[int] = set.union(*tmp_list)
                tmp_list: List[set] = [incoming.successors_right for intersection in intersections for incoming in
                                       intersection.incomings]
                successors_right: Set[int] = set.union(*tmp_list)
                all_successors = set.union(successors_straight, successors_right, successors_left)

        # select unique colors from colormap for each lanelet's center_line

        incoming_vertices_fill = list()
        crossing_vertices_fill = list()
        succ_left_paths = list()
        succ_straight_paths = list()
        succ_right_paths = list()

        vertices_fill = list()
        coordinates_left_border_vertices = np.empty((0, 2))
        coordinates_right_border_vertices = np.empty((0, 2))
        direction_list = list()
        center_paths = list()
        left_paths = list()
        right_paths = list()

        if draw_traffic_lights:
            center_line_color_dict = collect_center_line_colors(obj, traffic_lights, time_begin)

        cmap_lanelet = colormap_idx(len(lanelets))

        # collect paths for drawing
        for i_lanelet, lanelet in enumerate(lanelets):
            if isinstance(draw_lanlet_ids, list) and lanelet.lanelet_id not in draw_lanlet_ids:
                continue

            # left bound
            if draw_border_vertices or draw_left_bound:
                if draw_border_vertices:
                    coordinates_left_border_vertices = np.vstack(
                            (coordinates_left_border_vertices, lanelet.left_vertices))

                if draw_line_markings and lanelet.line_marking_left_vertices is not LineMarking.UNKNOWN and \
                        lanelet.line_marking_left_vertices is not LineMarking.NO_MARKING:
                    linestyle, dashes, linewidth_metres = line_marking_to_linestyle(lanelet.line_marking_left_vertices)
                    if lanelet.distance[-1] <= linewidth_metres:
                        left_paths.append(Path(lanelet.left_vertices, closed=False))
                    else:
                        tmp_left = lanelet.left_vertices.copy()
                        tmp_left[0, :] = lanelet.interpolate_position(linewidth_metres / 2)[2]
                        tmp_left[-1, :] = lanelet.interpolate_position(lanelet.distance[-1] - linewidth_metres / 2)[2]
                        line = LineDataUnits(tmp_left[:, 0], tmp_left[:, 1], zorder=ZOrders.LEFT_BOUND,
                                             linewidth=linewidth_metres, alpha=1.0, color=left_bound_color,
                                             linestyle=linestyle, dashes=dashes)
                        self.static_artists.append(line)
                else:
                    left_paths.append(Path(lanelet.left_vertices, closed=False))

            # right bound
            if draw_border_vertices or draw_right_bound:
                if draw_border_vertices:
                    coordinates_right_border_vertices = np.vstack(
                            (coordinates_right_border_vertices, lanelet.right_vertices))

                if draw_line_markings and lanelet.line_marking_right_vertices is not LineMarking.UNKNOWN and \
                        lanelet.line_marking_right_vertices is not LineMarking.NO_MARKING:
                    linestyle, dashes, linewidth_metres = line_marking_to_linestyle(lanelet.line_marking_right_vertices)
                    if lanelet.distance[-1] <= linewidth_metres:
                        right_paths.append(Path(lanelet.right_vertices, closed=False))
                    else:
                        tmp_right = lanelet.right_vertices.copy()
                        tmp_right[0, :] = lanelet.interpolate_position(linewidth_metres / 2)[1]
                        tmp_right[-1, :] = lanelet.interpolate_position(lanelet.distance[-1] - linewidth_metres / 2)[1]
                        line = LineDataUnits(tmp_right[:, 0], tmp_right[:, 1], zorder=ZOrders.RIGHT_BOUND,
                                             linewidth=linewidth_metres, alpha=1.0, color=right_bound_color,
                                             linestyle=linestyle, dashes=dashes)
                        self.static_artists.append(line)
                else:
                    right_paths.append(Path(lanelet.right_vertices, closed=False))

            # stop line
            if draw_stop_line and lanelet.stop_line:
                stop_line = np.vstack([lanelet.stop_line.start, lanelet.stop_line.end])
                linestyle, dashes, linewidth_metres = line_marking_to_linestyle(lanelet.stop_line.line_marking)
                # cut off in the beginning, because linewidth_metres is added
                # later
                vec = stop_line[1, :] - stop_line[0, :]
                tangent = vec / np.linalg.norm(vec)
                stop_line[0, :] += linewidth_metres * tangent / 2
                stop_line[1, :] -= linewidth_metres * tangent / 2
                line = LineDataUnits(stop_line[:, 0], stop_line[:, 1], zorder=ZOrders.STOP_LINE,
                                     linewidth=linewidth_metres, alpha=1.0, color=stop_line_color, linestyle=linestyle,
                                     dashes=dashes)
                self.static_artists.append(line)

            if unique_colors:
                # set center bound color to unique value
                center_bound_color = cmap_lanelet(i_lanelet)

            # direction arrow
            if draw_start_and_direction:
                center = lanelet.center_vertices[0]
                tan_vec = np.array(lanelet.right_vertices[0]) - np.array(lanelet.left_vertices[0])
                path = get_arrow_path_at(center[0], center[1], math.atan2(tan_vec[1], tan_vec[0]) + 0.5 * np.pi)
                if unique_colors:
                    direction_list.append(mpl.patches.PathPatch(path, color=center_bound_color, lw=0.5,
                                                                zorder=ZOrders.DIRECTION_ARROW,
                                                                antialiased=antialiased))
                else:
                    direction_list.append(path)

            # visualize traffic light state through colored center bound
            has_traffic_light = draw_traffic_lights and lanelet.lanelet_id in center_line_color_dict
            if has_traffic_light:
                light_state = center_line_color_dict[lanelet.lanelet_id]

                if light_state is not TrafficLightState.INACTIVE:
                    linewidth_metres = 0.75
                    # dashed line for red_yellow
                    linestyle = '--' if light_state == TrafficLightState.RED_YELLOW else '-'
                    dashes = (5, 5) if linestyle == '--' else (None, None)

                    # cut off in the beginning, because linewidth_metres is
                    # added
                    # later
                    tmp_center = lanelet.center_vertices.copy()
                    if lanelet.distance[-1] > linewidth_metres:
                        tmp_center[0, :] = lanelet.interpolate_position(linewidth_metres)[0]
                    zorder = ZOrders.LIGHT_STATE_GREEN if light_state == TrafficLightState.GREEN else \
                        ZOrders.LIGHT_STATE_OTHER
                    line = LineDataUnits(tmp_center[:, 0], tmp_center[:, 1], zorder=zorder, linewidth=linewidth_metres,
                                         alpha=0.7, color=traffic_light_color_dict(light_state, traffic_light_colors),
                                         linestyle=linestyle, dashes=dashes)
                    self.dynamic_artists.append(line)

            # draw colored center bound. Hierarchy or colors: successors > usual
            # center bound
            is_successor = draw_intersections and draw_successors and lanelet.lanelet_id in all_successors
            if is_successor:
                if lanelet.lanelet_id in successors_left:
                    succ_left_paths.append(Path(lanelet.center_vertices, closed=False))
                elif lanelet.lanelet_id in successors_straight:
                    succ_straight_paths.append(Path(lanelet.center_vertices, closed=False))
                else:
                    succ_right_paths.append(Path(lanelet.center_vertices, closed=False))

            elif draw_center_bound:
                if unique_colors:
                    center_paths.append(mpl.patches.PathPatch(Path(lanelet.center_vertices, closed=False),
                                                              edgecolor=center_bound_color, facecolor='none',
                                                              lw=draw_linewidth, zorder=ZOrders.CENTER_BOUND,
                                                              antialiased=antialiased))
                elif colormap_tangent:
                    relative_angle = draw_params['relative_angle']
                    points = lanelet.center_vertices.reshape(-1, 1, 2)
                    angles = get_tangent_angle(points[:, 0, :], relative_angle)
                    segments = np.concatenate([points[:-1], points[1:]], axis=1)
                    norm = plt.Normalize(0, 360)
                    lc = collections.LineCollection(segments, cmap='hsv', norm=norm, lw=draw_linewidth,
                                                    zorder=ZOrders.CENTER_BOUND, antialiased=antialiased)
                    lc.set_array(angles)
                    self.static_collections.append(lc)

            is_incoming_lanelet = draw_intersections and draw_incoming_lanelets and (
                    lanelet.lanelet_id in incoming_lanelets)
            is_crossing = draw_intersections and draw_crossings and (lanelet.lanelet_id in crossings)

            # Draw lanelet area
            if fill_lanelet:
                if not is_incoming_lanelet and not is_crossing:
                    vertices_fill.append(np.concatenate((lanelet.right_vertices, np.flip(lanelet.left_vertices, 0))))

            # collect incoming lanelets in separate list for plotting in
            # different color
            if is_incoming_lanelet:
                incoming_vertices_fill.append(
                        np.concatenate((lanelet.right_vertices, np.flip(lanelet.left_vertices, 0))))
            elif is_crossing:
                crossing_vertices_fill.append(
                        np.concatenate((lanelet.right_vertices, np.flip(lanelet.left_vertices, 0))))

            # Draw labels
            if show_label or show_intersection_labels or draw_traffic_signs:
                strings = []
                if show_label:
                    strings.append(str(lanelet.lanelet_id))
                if is_incoming_lanelet and show_intersection_labels:
                    strings.append(f'int_id: {inc_2_intersections[lanelet.lanelet_id].intersection_id}')
                    strings.append('inc_id: ' + str(incomings_id[lanelet.lanelet_id]))
                    strings.append('inc_left: ' + str(incomings_left[lanelet.lanelet_id]))
                if draw_traffic_signs and show_traffic_sign_label:
                    traffic_signs_tmp = [obj._traffic_signs[id] for id in lanelet.traffic_signs]
                    if traffic_signs_tmp:
                        # add as text to label
                        str_tmp = 'sign: '
                        add_str = ''
                        for sign in traffic_signs_tmp:
                            for el in sign.traffic_sign_elements:
                                # TrafficSignIDGermany(
                                # el.traffic_sign_element_id).name would give
                                # the
                                # name
                                str_tmp += add_str + el.traffic_sign_element_id.value
                                add_str = ', '

                        strings.append(str_tmp)

                label_string = ', '.join(strings)
                if len(label_string) > 0:
                    # compute normal angle of label box
                    clr_positions = lanelet.interpolate_position(0.5 * lanelet.distance[-1])
                    normal_vector = np.array(clr_positions[1]) - np.array(clr_positions[2])
                    angle = np.rad2deg(math.atan2(normal_vector[1], normal_vector[0])) - 90
                    angle = angle if Interval(-90, 90).contains(angle) else angle - 180

                    self.static_artists.append(text.Text(clr_positions[0][0], clr_positions[0][1], label_string,
                                                         bbox={'facecolor': center_bound_color, 'pad': 2},
                                                         horizontalalignment='center', verticalalignment='center',
                                                         rotation=angle, zorder=ZOrders.LANELET_LABEL))

        # draw paths and collect axis handles
        if draw_right_bound:
            self.static_collections.append(
                    collections.PathCollection(right_paths, edgecolor=right_bound_color, facecolor='none',
                                               lw=draw_linewidth, zorder=ZOrders.RIGHT_BOUND, antialiased=antialiased))
        if draw_left_bound:
            self.static_collections.append(
                    collections.PathCollection(left_paths, edgecolor=left_bound_color, facecolor='none',
                                               lw=draw_linewidth, zorder=ZOrders.LEFT_BOUND, antialiased=antialiased))
        if unique_colors:
            if draw_center_bound:
                if draw_center_bound:
                    self.static_collections.append(
                            collections.PatchCollection(center_paths, match_original=True, zorder=ZOrders.CENTER_BOUND,
                                                        antialiased=antialiased))
                if draw_start_and_direction:
                    self.static_collections.append(collections.PatchCollection(direction_list, match_original=True,
                                                                               zorder=ZOrders.DIRECTION_ARROW,
                                                                               antialiased=antialiased))

        elif not colormap_tangent:
            if draw_center_bound:
                self.static_collections.append(
                        collections.PathCollection(center_paths, edgecolor=center_bound_color, facecolor='none',
                                                   lw=draw_linewidth, zorder=ZOrders.CENTER_BOUND,
                                                   antialiased=antialiased))
            if draw_start_and_direction:
                self.static_collections.append(
                        collections.PathCollection(direction_list, color=center_bound_color, lw=0.5,
                                                   zorder=ZOrders.DIRECTION_ARROW, antialiased=antialiased))

        if successors_left:
            self.static_collections.append(
                    collections.PathCollection(succ_left_paths, edgecolor=successors_left_color, facecolor='none',
                                               lw=draw_linewidth * 3.0, zorder=ZOrders.SUCCESSORS,
                                               antialiased=antialiased))
        if successors_straight:
            self.static_collections.append(
                    collections.PathCollection(succ_straight_paths, edgecolor=successors_straight_color,
                                               facecolor='none', lw=draw_linewidth * 3.0, zorder=ZOrders.SUCCESSORS,
                                               antialiased=antialiased))
        if successors_right:
            self.static_collections.append(
                    collections.PathCollection(succ_right_paths, edgecolor=successors_right_color, facecolor='none',
                                               lw=draw_linewidth * 3.0, zorder=ZOrders.SUCCESSORS,
                                               antialiased=antialiased))

        # fill lanelets with facecolor
        self.static_collections.append(
                collections.PolyCollection(vertices_fill, transOffset=self.ax.transData, zorder=ZOrders.LANELET_POLY,
                                           facecolor=facecolor, edgecolor='none', antialiased=antialiased))
        if incoming_vertices_fill:
            self.static_collections.append(
                    collections.PolyCollection(incoming_vertices_fill, transOffset=self.ax.transData,
                                               facecolor=incoming_lanelets_color, edgecolor='none',
                                               zorder=ZOrders.INCOMING_POLY, antialiased=antialiased))
        if crossing_vertices_fill:
            self.static_collections.append(
                    collections.PolyCollection(crossing_vertices_fill, transOffset=self.ax.transData,
                                               facecolor=crossings_color, edgecolor='none',
                                               zorder=ZOrders.CROSSING_POLY, antialiased=antialiased))

        # draw_border_vertices
        if draw_border_vertices:
            # left vertices
            self.static_collections.append(
                    collections.EllipseCollection(np.ones([coordinates_left_border_vertices.shape[0], 1]) * 1.5,
                                                  np.ones([coordinates_left_border_vertices.shape[0], 1]) * 1.5,
                                                  np.zeros([coordinates_left_border_vertices.shape[0], 1]),
                                                  offsets=coordinates_left_border_vertices, color=left_bound_color,
                                                  transOffset=self.ax.transData,
                                                  zorder=ZOrders.LEFT_BOUND + 0.1,))

            # right_vertices
            self.static_collections.append(
                    collections.EllipseCollection(np.ones([coordinates_right_border_vertices.shape[0], 1]) * 1.5,
                                                  np.ones([coordinates_right_border_vertices.shape[0], 1]) * 1.5,
                                                  np.zeros([coordinates_right_border_vertices.shape[0], 1]),
                                                  offsets=coordinates_right_border_vertices, color=right_bound_color,
                                                  transOffset=self.ax.transData,
                                                  zorder=ZOrders.LEFT_BOUND + 0.1,))

        if draw_traffic_signs:
            # draw actual traffic sign
            for sign in traffic_signs:
                sign.draw(self, draw_params, call_stack)

        if draw_traffic_lights:
            # draw actual traffic light
            for light in traffic_lights:
                light.draw(self, draw_params, call_stack)

    def draw_planning_problem_set(self, obj: PlanningProblemSet, draw_params: Union[ParamServer, dict, None],
                                  call_stack: Tuple[str, ...]) -> None:
        """
        Draws all or selected planning problems from the planning problem set. Planning problems can be selected by
        providing IDs in`drawing_params[planning_problem_set][draw_ids]`

        :param obj: object to be plotted
        :param draw_params: parameters for plotting given by a nested dict
            that recreates the structure of an object or a ParamServer object
        :param call_stack: tuple of string containing the call stack,
            which allows for differentiation of plotting styles depending on the call stack
        :return: None
        """
        draw_params = self._get_draw_params(draw_params)
        call_stack = tuple(list(call_stack) + ['planning_problem_set'])
        draw_ids = draw_params.by_callstack(call_stack, 'draw_ids')

        for pp_id, problem in obj.planning_problem_dict.items():
            if draw_ids == 'all' or pp_id in draw_ids:
                self.draw_planning_problem(problem, draw_params, call_stack)

    def draw_planning_problem(self, obj: PlanningProblem, draw_params: Union[ParamServer, dict, None],
                              call_stack: Tuple[str, ...]) -> None:
        """
        Draw initial state and goal region of the planning problem

        :param obj: object to be plotted
        :param draw_params: parameters for plotting given by a nested dict
            that recreates the structure of an object or a ParamServer object
        :param call_stack: tuple of string containing the call stack,
            which allows for differentiation of plotting styles depending on the call stack
        :return: None
        """
        draw_params = self._get_draw_params(draw_params)
        call_stack = tuple(list(call_stack) + ['planning_problem'])
        self.draw_initital_state(obj.initial_state, draw_params, call_stack)
        self.draw_goal_region(obj.goal, draw_params, call_stack)

    def draw_initital_state(self, obj: State, draw_params: Union[ParamServer, dict, None],
                            call_stack: Tuple[str, ...]) -> None:
        """
        Draw initial state with label

        :param obj: object to be plotted
        :param draw_params: parameters for plotting given by a nested dict
            that recreates the structure of an object or a ParamServer object
        :param call_stack: tuple of string containing the call stack,
            which allows for differentiation of plotting styles depending on the call stack
        :return: None
        """
        draw_params = self._get_draw_params(draw_params)
        call_stack = call_stack + ('initial_state',)
        zorder = draw_params.by_callstack(call_stack, 'label_zorder')
        label = draw_params.by_callstack(call_stack, 'label')

        obj.draw(self, draw_params, call_stack)
        self.static_artists.append(
                text.Annotation(label, xy=(obj.position[0] + 1, obj.position[1]), textcoords='data', zorder=zorder))

    def draw_goal_region(self, obj: GoalRegion, draw_params: Union[ParamServer, dict, None],
                         call_stack: Tuple[str, ...]) -> None:
        """
        Draw goal states from goal region

        :param obj: object to be plotted
        :param draw_params: parameters for plotting given by a nested dict
            that recreates the structure of an object or a ParamServer object
        :param call_stack: tuple of string containing the call stack,
            which allows for differentiation of plotting styles depending on the call stack
        :return: None
        """
        draw_params = self._get_draw_params(draw_params)
        if call_stack == ():
            call_stack = tuple(['planning_problem_set'])
        call_stack = tuple(list(call_stack) + ['goal_region'])
        for goal_state in obj.state_list:
            self.draw_goal_state(goal_state, draw_params, call_stack)

    def draw_goal_state(self, obj: State, draw_params: Union[ParamServer, dict, None],
                        call_stack: Tuple[str, ...]) -> None:
        """
        Draw goal states

        :param obj: object to be plotted
        :param draw_params: parameters for plotting given by a nested dict
            that recreates the structure of an object or a ParamServer object
        :param call_stack: tuple of string containing the call stack,
            which allows for differentiation of plotting styles depending on the call stack
        :return: None
        """
        draw_params = self._get_draw_params(draw_params)
        if hasattr(obj, 'position'):
            if type(obj.position) == list:
                for pos in obj.position:
                    pos.draw(self, draw_params, call_stack)
            else:
                obj.position.draw(self, draw_params, call_stack)

    def draw_traffic_light_sign(self, obj: Union[TrafficLight, TrafficSign],
                                draw_params: Union[ParamServer, dict, None], call_stack: Tuple[str, ...]):
        """
        Draw traffic sings and lights

        :param obj: object to be plotted
        :param draw_params: parameters for plotting given by a nested dict
            that recreates the structure of an object or a ParamServer object
        :param call_stack: tuple of string containing the call stack,
            which allows for differentiation of plotting styles depending on the call stack
        :return: None
        """
        draw_params = self._get_draw_params(draw_params)
        # traffic signs and lights have to be collected and drawn only right
        # before rendering to allow correct grouping
        self.traffic_sign_call_stack = call_stack
        self.traffic_sign_draw_params = draw_params
        self.traffic_signs.append(obj)

    def _draw_signal_state(self, sig: SignalState, occ: Occupancy, draw_params: Union[ParamServer, dict, None],
                           call_stack: Tuple[str, ...]):
        """
        Draw the vehicle signals

        :param sig: signal state to be drawn
        :param occ: Occupancy at current time step
        :param draw_params: parameters for plotting given by a nested dict
            that recreates the structure of an object or a ParamServer object
        :param call_stack: tuple of string containing the call stack, which allows for differentiation of plotting
            styles depending on the call stack
        :return: None
        """
        draw_params = self._get_draw_params(draw_params)
        call_stack = call_stack + ('signals',)
        signal_radius = draw_params.by_callstack(call_stack, 'signal_radius')

        indicators = []
        braking = []

        # indicators
        if isinstance(occ.shape, Rectangle):
            if hasattr(sig, 'hazard_warning_lights') and sig.hazard_warning_lights is True:
                indicators.extend(
                        [occ.shape.vertices[0], occ.shape.vertices[1], occ.shape.vertices[2], occ.shape.vertices[3]])
            else:
                if hasattr(sig, 'indicator_left') and sig.indicator_left is True:
                    indicators.extend([occ.shape.vertices[1], occ.shape.vertices[2]])
                if hasattr(sig, 'indicator_right') and sig.indicator_right is True:
                    indicators.extend([occ.shape.vertices[0], occ.shape.vertices[3]])

            for e in indicators:
                self.draw_ellipse(e, signal_radius, signal_radius, draw_params, call_stack + ('indicator',))

            # braking lights
            if hasattr(sig, 'braking_lights') and sig.braking_lights is True:
                braking.extend([occ.shape.vertices[0], occ.shape.vertices[1]])

            for e in braking:
                self.draw_ellipse(e, signal_radius * 1.5, signal_radius * 1.5, draw_params, call_stack + ('braking',))

            # blue lights
            if hasattr(sig, 'flashing_blue_lights') and sig.flashing_blue_lights is True:
                pos = occ.shape.center
                self.draw_ellipse(pos, signal_radius, signal_radius, draw_params, call_stack + ('bluelight',))

            # horn
            if hasattr(sig, 'horn') and sig.horn is True:
                pos = occ.shape.center
                self.draw_ellipse(pos, signal_radius * 1.5, signal_radius * 1.5, draw_params, call_stack + ('horn',))

        else:
            warnings.warn('Plotting signal states only implemented for '
                          'obstacle_shapes Rectangle.')



import copy
import enum
import os
from collections import defaultdict
from typing import Dict
from scipy.cluster.hierarchy import linkage, fcluster

from PIL import Image
from commonroad.geometry.shape import *
from commonroad.scenario.traffic_sign import TrafficSign, \
    TrafficLight, \
    TrafficSignIDUsa
from commonroad.visualization.param_server import ParamServer
from matplotlib.axes import Axes, mtext
from matplotlib.offsetbox import OffsetImage, \
    AnnotationBbox, \
    HPacker, \
    TextArea, \
    VPacker, \
    OffsetBox

# path to traffic sign .png files
traffic_sign_path = os.path.join(os.path.dirname(__file__), 'traffic_signs/')

speed_limit_factors = {'mph': 2.23694, 'kmh': 3.6, 'ms': 1.0}

# default scaling for traffic sign images and annotations
px_per_metre = 0.018
px_per_metre_text = 0.078


class TextAreaAutoscale(TextArea):
    def __init__(self, s,
                 px_per_metre=px_per_metre_text,
                 textprops=None,
                 multilinebaseline=None,
                 minimumdescent=True,
                 ):
        super().__init__(s,
                         textprops=textprops,
                         multilinebaseline=multilinebaseline,
                         minimumdescent=minimumdescent)

        self.px_per_metre = px_per_metre
        textprops_init = copy.copy(textprops)
        if textprops_init is None:
            textprops_init = {}
        textprops_init.setdefault("va", "center")
        textprops_init.setdefault("ha", "center")
        self.textprops_init = textprops_init
        self.dx_m = None
        self.dy_m = None
        self.dx_pix = None
        self.dy_pix = None

    def set_ax_lims(self, dx_m, dy_m, dx_pix, dy_pix):
        self.dx_m = dx_m
        self.dy_m = dy_m
        self.dx_pix = dx_pix
        self.dy_pix = dy_pix

    def ax_update(self, ax: Axes):
        '''
        Update image size based on axes limits and window size in pixels
        '''
        bbox = ax.get_window_extent()
        width_px, height_px = bbox.width, bbox.height
        # Get the range for the new area
        _, _, xdelta, ydelta = ax.viewLim.bounds
        self.set_ax_lims(xdelta, ydelta, width_px, height_px)

    def get_extent(self, renderer):
        correction_x = self.dx_pix / self.dx_m * self.px_per_metre
        correction_y = self.dy_pix / self.dy_m * self.px_per_metre
        corr = min(correction_x, correction_y)
        if 'size' in self.textprops_init:
            textprops = copy.copy(self.textprops_init)
            textprops['size'] *= corr
            f = self._text.figure
            self._text = mtext.Text(0, 0, self.get_text(), **textprops)
            self._text.figure = f
            self._children = [self._text]
            self._text.set_transform(self.offset_transform +
                                     self._baseline_transform)
        _, h_, d_ = renderer.get_text_width_height_descent(
            "lp", self._text._fontproperties, ismath=False)

        bbox, info, d = self._text._get_layout(renderer)
        w, h = bbox.width, bbox.height

        self._baseline_transform.clear()

        if len(info) > 1 and self._multilinebaseline:
            d_new = 0.5 * h - 0.5 * (h_ - d_)
            self._baseline_transform.translate(0, d - d_new)
            d = d_new

        else:  # single line

            h_d = max(h_ - d_, h - d)

            if self.get_minimumdescent():
                ## to have a minimum descent, #i.e., "l" and "p" have same
                ## descents.
                d = max(d, d_)

            h = h_d + d

        return w * corr, h * corr, 0., d * corr


class OffsetImageAutoscale(OffsetImage):
    """OffsetImage that scales proportionally to data units (not display units, i.e. pixels)
    and includes overlayed text."""

    def __init__(self, arr,
                 px_per_metre=px_per_metre,
                 px_per_metre_text=px_per_metre_text,
                 text=None,
                 txt_offset_x=0.7,
                 txt_offset_y=0.45,
                 textprops=None,
                 cmap=None,
                 norm=None,
                 interpolation=None,
                 origin=None,
                 filternorm=True,
                 filterrad=4.0,
                 resample=False,
                 dpi_cor=True,
                 **kwargs
                 ):
        super().__init__(arr,
                         cmap=cmap,
                         norm=norm,
                         interpolation=interpolation,
                         origin=origin,
                         filternorm=filternorm,
                         filterrad=filterrad,
                         resample=resample,
                         dpi_cor=dpi_cor,
                         **kwargs)

        self.text_area = TextAreaAutoscale(text, px_per_metre=px_per_metre_text, textprops=textprops)
        self.txt_offset_x = txt_offset_x
        self.txt_offset_y = txt_offset_y
        self.px_per_metre = px_per_metre
        self.dx_m = None
        self.dy_m = None
        self.dx_pix = None
        self.dy_pix = None

    def set_ax_lims(self, dx_m, dy_m, dx_pix, dy_pix):
        self.dx_m = dx_m
        self.dy_m = dy_m
        self.dx_pix = dx_pix
        self.dy_pix = dy_pix

    def ax_update(self, ax: Axes):
        '''
        Update image size based on axes limits and window size in pixels
        '''
        bbox = ax.get_window_extent()
        width_px, height_px = bbox.width, bbox.height
        # Get the range for the new area
        _, _, xdelta, ydelta = ax.viewLim.bounds
        self.set_ax_lims(xdelta, ydelta, width_px, height_px)
        self.text_area.set_ax_lims(xdelta, ydelta, width_px, height_px)

    def get_extent(self, renderer=None):
        self.text_area._text.figure = self.figure
        if renderer is not None:
            self.text_area.get_extent(renderer)

        if renderer is not None and self._dpi_cor:  # True, do correction
            dpi_cor = renderer.points_to_pixels(1.)
        else:
            dpi_cor = 1.

        data = self.get_data()
        correction_x = self.dx_pix / self.dx_m * self.px_per_metre
        correction_y = self.dy_pix / self.dy_m * self.px_per_metre
        corr = min(correction_x, correction_y)
        ny, nx = data.shape[:2]
        w, h = dpi_cor * nx * corr, dpi_cor * ny * corr

        return w, h, 0, 0

    def draw(self, renderer):
        # docstring inherited
        self.image.draw(renderer)
        self.text_area._text.figure = self.figure
        self.text_area._text.draw(renderer)
        self.stale = False

    def set_offset(self, xy):
        super().set_offset(xy)
        self.text_area.offset_transform.clear()
        w, h, _, _ = self.get_extent()
        xy = (xy[0] + w * self.txt_offset_x, xy[1] + h * self.txt_offset_y)
        self.text_area.offset_transform.translate(xy[0], xy[1])


def isfloat(value: str):
    try:
        float(value)
        return True
    except ValueError:
        return False


def speed_limit_factor(country_code) -> float:
    """Determine factor for speed_limit_unit by country code."""
    # dicts for units other than kph
    mph_countries = [TrafficSignIDUsa]

    if type(country_code) in mph_countries:
        return speed_limit_factors['mph']
    else:
        return speed_limit_factors['kmh']


# denotes traffic signs that are speed limits
unit_conversion_required = ['274', '274.1', '275', '1004-31', 'R2-1', 'r301']


def text_prop_dict() -> dict:
    """Properties of text for additional_value.
    mpl_args: text properties
    rescale_threshold: max. length of additional value string. Above this length, the font size is scaled down
    position_offset_y: vertical offset of additional text measured from the bottom of the traffic sign
        (proportional to traffic sign size)
    """
    return {
        'default': {
            'mpl_args': {
                'weight': 'bold', 'size': 13.5
            },
            'rescale_threshold': 2,
            'position_offset_y': 0.45
        },
        '262': {
            'mpl_args': {
                'weight': 'bold', 'size': 8
            },
            'rescale_threshold': 2,
            'position_offset_y': 0.53
        },
        '264': {
            'mpl_args': {
                'weight': 'bold', 'size': 12
            },
            'rescale_threshold': 2,
            'position_offset_y': 0.47
        },
        '265': {
            'mpl_args': {
                'weight': 'normal', 'size': 13
            },
            'rescale_threshold': 2,
            'position_offset_y': 0.45
        },
        '266': {
            'mpl_args': {
                'weight': 'bold', 'size': 5
            },
            'rescale_threshold': 2,
            'position_offset_y': 0.37
        },
        '274': {
            'mpl_args': {
                'weight': 'bold', 'size': 13.5
            },
            'rescale_threshold': 2,
            'position_offset_y': 0.45
        },
        '274.1': {
            'mpl_args': {
                'weight': 'bold', 'size': 8.5
            },
            'rescale_threshold': 2,
            'position_offset_y': 0.65
        },
        '275': {
            'mpl_args': {
                'weight': 'bold', 'color': 'white', 'size': 13.5
            }, 'rescale_threshold': 2,
            'position_offset_y': 0.45
        },
        '278': {
            'mpl_args': {
                'weight': 'bold', 'color': 'gray', 'size': 10
            },
            'position_offset': -16.5,
            'position_offset_y': 0.45
        },
        '279': {
            'mpl_args': {
                'weight': 'bold', 'color': 'white', 'size': 10
            },
            'position_offset': -16.5,
            'position_offset_y': 0.45
        },
        '310': {
            'mpl_args': {
                'weight': 'normal', 'color': 'black', 'size': 10,
            },
            'position_offset_y': -0.55
        },
        '380': {
            'mpl_args': {
                'weight': 'bold', 'color': 'white', 'size': 10
            },
            'position_offset_y': 0.45
        },
        '381': {
            'mpl_args': {
                'weight': 'bold', 'color': 'white', 'size': 10
            },
            'position_offset_y': 0.45
        },
        '1040-30': {
            'mpl_args': {
                'weight': 'normal', 'color': 'black', 'size': 8
            },
            'position_offset_y': 0.4
        },
        '1001-30': {
            'mpl_args': {
                'weight': 'normal', 'color': 'black', 'size': 8
            },
            'position_offset_y': 0.4
        },
        '1004-31': {
            'mpl_args': {
                'weight': 'normal', 'color': 'black', 'size': 8
            },
            'position_offset_y': 0.4
        },
        'R2-1': {
            'mpl_args': {
                'weight': 'normal', 'color': 'black', 'size': 10.5
            },
            'position_offset_y': 0.3
        },
        'r301': {
            'mpl_args': {
                'weight': 'bold', 'size': 13.5
            },
            'rescale_threshold': 2,
            'position_offset_y': 0.45
        },
    }


def rescale_text(string: str, prop: dict, scale_factor: float,
                 default_scale_factor: float) -> dict:
    """Rescales text size proportionally to the max. number of strings given
    by prop['rescale_threshold'] and to the
    'scale_factor' compared to the default scale_factor. Used e.g. for
    fitting speed limits into the traffic sign."""
    prop = copy.deepcopy(prop)
    if default_scale_factor != scale_factor:
        tmp_scale_factor = scale_factor / default_scale_factor
        if 'position_offset' in prop:
            prop['position_offset'] *= tmp_scale_factor

        if 'mpl_args' in prop and 'size' in prop['mpl_args']:
            prop['mpl_args']['size'] *= tmp_scale_factor

    if 'rescale_threshold' in prop:
        if len(string) > prop['rescale_threshold']:
            if 'mpl_args' in prop and 'size' in prop['mpl_args']:
                prop['mpl_args']['size'] *= prop['rescale_threshold'] / len(
                        string) * 1.1
            if 'position_offset' in prop:
                prop['position_offset'] *= prop['rescale_threshold'] / len(
                        string) * 1.35

    return prop


def create_img_boxes_traffic_sign(
        traffic_signs: Union[List[TrafficSign], TrafficSign],
        draw_params: ParamServer, call_stack: Tuple[str, ...], rnd) -> Dict[
    Tuple[float, float], List[OffsetBox]]:
    """
    For each Traffic sign an OffsetBox is created, containing the png image
    and optionally labels. These boxes can
    be stacked horizontally later when multiple signs or lights share the
    same position.
    :param traffic_signs:
    :param draw_params:
    :param call_stack:
    :param rnd: MPRenderer
    :return:
    """
    if type(traffic_signs) is not list:
        traffic_signs = [traffic_signs]

    if len(traffic_signs) == 0:
        return dict()

    scale_factor = draw_params.by_callstack(call_stack,
                                            ('traffic_sign', 'scale_factor'))
    speed_limit_unit = draw_params.by_callstack(call_stack, (
            'traffic_sign', 'speed_limit_unit'))
    show_label_default = draw_params.by_callstack(call_stack, (
            'traffic_sign', 'show_label'))
    show_traffic_signs = draw_params.by_callstack(call_stack, (
            'traffic_sign', 'show_traffic_signs'))
    zorder = draw_params.by_callstack(call_stack, ('traffic_sign', 'zorder'))

    scale_factor_default = draw_params.by_callstack(call_stack, (
            'traffic_sign', 'scale_factor'))

    assert any([show_traffic_signs == 'all',
                isinstance(show_traffic_signs, list) and type(
                        show_traffic_signs[0] is enum)]), 'Plotting option ' \
                                                          'traffic_sign.show_traffic_signs must ' \
                                                          'be either "all" or ' \
                                                          '' \
                                                          '' \
                                                          'list of type ' \
                                                          'TrafficSignID'

    prop_dict = text_prop_dict()
    imageboxes_all = defaultdict(list)

    for traffic_sign in traffic_signs:
        if traffic_sign.virtual is True or traffic_sign.position is None:
            continue
        imageboxes = []
        for element in traffic_sign.traffic_sign_elements:
            el_id = element.traffic_sign_element_id
            if not (show_traffic_signs == 'all' or el_id in show_traffic_signs):
                continue
            show_label = show_label_default
            path = os.path.join(traffic_sign_path, el_id.__class__.__name__,
                                el_id.value + '.png')
            plot_img = True
            # get png image
            if not os.path.exists(path):
                warnings.warn(f"File for traffic sign {element} at {path} does not exist!")
                path = os.path.join(traffic_sign_path, '.png')
                if not os.path.exists(path):
                    show_label = True
                    plot_img = False

            boxes = []  # collect matplotlib offset boxes for text and images
            if show_label:
                boxes.append(TextArea(el_id.value))

            if plot_img:
                # plot traffic sign
                sign_img = Image.open(path)
                if len(element.additional_values) > 0:
                    if element.traffic_sign_element_id.value in unit_conversion_required \
                            and isfloat(element.additional_values[0]):
                        if speed_limit_unit == 'auto':
                            speed_factor = speed_limit_factor(
                                element.traffic_sign_element_id)
                        else:
                            speed_factor = speed_limit_factors[speed_limit_unit]

                        add_text = str(round(
                            speed_factor * float(element.additional_values[0])))
                    else:
                        add_text = '\n'.join(element.additional_values)

                    props = prop_dict[el_id.value] if el_id.value in prop_dict else prop_dict['default']
                    props = rescale_text(add_text, props, scale_factor,
                                         scale_factor_default)
                    props_txt = props['mpl_args']
                    txt_offset_y = props['position_offset_y'] if 'position_offset_y' in props else -0.2
                else:
                    txt_offset_y = -0.2
                    props_txt = None
                    add_text = None

                if sign_img.mode != "RGBA":
                    sign_img = sign_img.convert("RGBA")

                boxes.append(
                    OffsetImageAutoscale(sign_img, text=add_text,
                                         px_per_metre=px_per_metre*scale_factor,
                                         px_per_metre_text=px_per_metre_text*scale_factor,
                                         txt_offset_y=txt_offset_y,
                                         textprops=props_txt,
                                         zorder=zorder,
                                         resample=True))
                # add callback for automatic rescaling of image
                rnd.add_callback('xlim_changed', boxes[-1].ax_update)

            # already stack label and img in case both are shown (prevents
            # misalignment with additional text)
            if len(boxes) > 1:
                boxes = [VPacker(children=boxes, pad=0, sep=0, align='center')]

            sep = 0
            # get additional values string (like speed limits) in case the image is not plotted
            if not plot_img:
                if element.traffic_sign_element_id.value in unit_conversion_required \
                        and isfloat(
                        element.additional_values[0]):
                    if speed_limit_unit == 'auto':
                        speed_factor = speed_limit_factor(
                                element.traffic_sign_element_id)
                    else:
                        speed_factor = speed_limit_factors[speed_limit_unit]

                    add_text = str(round(
                            speed_factor * float(element.additional_values[0])))
                else:
                    add_text = '\n'.join(element.additional_values)

                props = prop_dict[
                    el_id.value] if el_id.value in prop_dict else {
                        'mpl_args': {}
                }
                props = rescale_text(add_text, props, scale_factor,
                                     scale_factor_default)
                boxes.append(TextAreaAutoscale(add_text, px_per_metre=px_per_metre_text * scale_factor,
                                               textprops=props['mpl_args']))
                # add callback for automatic rescaling of text
                rnd.add_callback('xlim_changed', boxes[-1].ax_update)

            # stack boxes vertically
            img = VPacker(children=boxes, pad=0, sep=sep, align='center')
            imageboxes.append(img)

        # horizontally stack all traffic sign elements of the traffic sign
        if len(imageboxes) > 0:
            hbox = HPacker(children=imageboxes, pad=0, sep=0.05,
                           align='baseline')
            imageboxes_all[tuple(traffic_sign.position.tolist())].append(hbox)

    return imageboxes_all


def create_img_boxes_traffic_lights(
        traffic_lights: Union[List[TrafficLight], TrafficLight],
        draw_params: ParamServer, call_stack: Tuple[str, ...], rnd) -> Dict[
    Tuple[float, float], List[OffsetBox]]:
    """
    For each Traffic light an OffsetBox is created, containing the png image
    and optionally labels. These boxes can
    be stacked horizontally later when multiple signs or lights share the
    same position.
    :param traffic_lights:
    :param draw_params:
    :param call_stack:
    :return:
    """
    if type(traffic_lights) is not list:
        traffic_lights = [traffic_lights]

    if len(traffic_lights) == 0:
        return dict()

    time_begin = draw_params.by_callstack(call_stack, ('time_begin',))
    scale_factor = draw_params.by_callstack(call_stack,
                                            ('traffic_light', 'scale_factor'))
    show_label = draw_params.by_callstack(call_stack,
                                          ('traffic_light', 'show_label'))
    zorder = draw_params.by_callstack(call_stack, ('traffic_light', 'zorder'))

    # plots all group members horizontally stacked
    imageboxes_all = defaultdict(list)
    for traffic_light in traffic_lights:
        if traffic_light.position is None:
            continue
        if traffic_light.active:
            state = traffic_light.get_state_at_time_step(time_begin)
            path = os.path.join(traffic_sign_path, 'traffic_light_state_' + str(
                    state.value) + '.png')
        else:
            path = os.path.join(traffic_sign_path,
                                'traffic_light_state_inactive.png')

        boxes = []  # collect matplotlib offset boxes for text and images
        sign_img = Image.open(path)
        boxes.append(OffsetImageAutoscale(sign_img,
                                          px_per_metre=px_per_metre * scale_factor,
                                          zorder=zorder,
                                          resample=True))
        rnd.add_callback('xlim_changed', boxes[-1].ax_update)

        if show_label:
            boxes.append(TextArea(str(state.value)))

        # stack boxes vertically
        img_box = VPacker(children=boxes, pad=0, sep=0, align='center')

        imageboxes_all[tuple(traffic_light.position.tolist())].append(img_box)

    return imageboxes_all


def draw_traffic_light_signs(traffic_lights_signs: Union[
    List[Union[TrafficLight, TrafficSign]], TrafficLight, TrafficSign],
                             draw_params: ParamServer,
                             call_stack: Tuple[str, ...],
                             rnd):
    """
    Draws OffsetBoxes which are first collected for all traffic signs and
    -lights. Boxes are stacked together when they
    share the same position.
    :param traffic_lights_signs:
    :param plot_limits:
    :param ax:
    :param draw_params:
    :param draw_func:
    :param handles:
    :param call_stack:
    :param rnd: MPRenderer
    :return:
    """
    kwargs = draw_params.by_callstack(call_stack, (
    'lanelet_network', 'kwargs_traffic_light_signs'))

    zorder_0 = draw_params.by_callstack(call_stack, ('traffic_light', 'zorder'))

    zorder_1 = draw_params.by_callstack(call_stack, ('traffic_sign', 'zorder'))

    zorder = min(zorder_0, zorder_1)
    threshold_grouping = 0.8  # [m] distance threshold for grouping traffic
    # light and/or signs

    if not isinstance(traffic_lights_signs, list):
        traffic_lights_signs = [traffic_lights_signs]

    traffic_signs = []
    traffic_lights = []

    for obj in traffic_lights_signs:
        if isinstance(obj, TrafficSign):
            traffic_signs.append(obj)
        elif isinstance(obj, TrafficLight):
            traffic_lights.append(obj)
        else:
            warnings.warn('Object of type {}, but expected type TrafficSign or '
                          'TrafficLight'.format(type(obj)))

    # collect ImageBoxes of traffic signs/lights grouped by their positions
    boxes_tl = create_img_boxes_traffic_lights(traffic_lights, draw_params,
                                               call_stack, rnd)
    boxes_signs = create_img_boxes_traffic_sign(traffic_signs, draw_params,
                                                call_stack, rnd)

    img_boxes = defaultdict(list)  # {position: List[OffsetBox]}
    [img_boxes[pos].extend(box_list) for pos, box_list in boxes_tl.items()]
    [img_boxes[pos].extend(box_list) for pos, box_list in boxes_signs.items()]

    if not img_boxes:
        return []

    positions = list(img_boxes.keys())
    box_lists = list(img_boxes.values())

    # group objects based on their positions' distances
    group_boxes = defaultdict(list)
    group_positions = defaultdict(list)
    groups = defaultdict(list)

    # find clusters where minimal pairwise distance is below distance threshold_grouping
    if len(positions) <= 1:
        groups[positions[0]] = box_lists[0]
    else:
        Z = linkage(np.array(positions), 'single', metric='chebyshev')
        clusters = fcluster(Z, threshold_grouping, criterion='distance')
        for i, cluster_id in enumerate(clusters):
            group_boxes[cluster_id].extend(box_lists[i])
            group_positions[cluster_id].append(positions[i])

        for cluster_id, boxes in group_boxes.items():
            groups[tuple(np.average(group_positions[cluster_id], axis=0).tolist())] = boxes

    # add default AnnotationBox args if not specified by user
    default_params = dict(xycoords='data', frameon=False)
    for param, value in default_params.items():
        if param not in kwargs:
            kwargs[param] = value

    artists = []
    # stack imageboxes of each group and draw
    for position_tmp, box_list_tmp in groups.items():
        position_tmp = np.array(position_tmp)
        kwargs_tmp = copy.deepcopy(kwargs)
        if 'xybox' not in kwargs_tmp:
            kwargs_tmp['xybox'] = position_tmp

        hbox = HPacker(children=box_list_tmp, pad=0, sep=0.1, align='baseline')
        ab = AnnotationBbox(hbox, position_tmp, **kwargs_tmp)
        ab.zorder = zorder
        artists.append(ab)
    return artists
